# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ("iter_iuse_vars",)


def iter_iuse_vars(env):
    """
    Iterate over (key, value) pairs of profile variables that contribute
    to implicit IUSE for EAPI 5 and later.

    @param env: Ebuild environment
    @type env: Mapping
    @rtype: iterator
    @return: iterator over (key, value) pairs of profile variables
    """

    for k in (
        "IUSE_IMPLICIT",
        "USE_EXPAND_IMPLICIT",
        "USE_EXPAND_UNPREFIXED",
        "USE_EXPAND",
    ):
        v = env.get(k)
        if v is not None:
            yield (k, v)

    use_expand_implicit = frozenset(env.get("USE_EXPAND_IMPLICIT", "").split())

    for v in (
        env.get("USE_EXPAND_UNPREFIXED", "").split() + env.get("USE_EXPAND", "").split()
    ):
        if v in use_expand_implicit:
            k = "USE_EXPAND_VALUES_" + v
            v = env.get(k)
            if v is not None:
                yield (k, v)



# test_get_operator.py -- Portage Unit Testing Functionality
# Copyright 2007 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import get_operator


class GetOperator(TestCase):
    def testGetOperator(self):

        # get_operator does not validate operators
        tests = [
            ("~", "~"),
            ("=", "="),
            (">", ">"),
            (">=", ">="),
            ("<=", "<="),
        ]

        test_cpvs = ["sys-apps/portage-2.1"]
        slots = [None, "1", "linux-2.5.6"]
        for cpv in test_cpvs:
            for test in tests:
                for slot in slots:
                    atom = cpv[:]
                    if slot:
                        atom += ":" + slot
                    result = get_operator(test[0] + atom)
                    self.assertEqual(
                        result,
                        test[1],
                        msg="get_operator({}) != {}".format(test[0] + atom, test[1]),
                    )

        result = get_operator("sys-apps/portage")
        self.assertEqual(result, None)

        result = get_operator("=sys-apps/portage-2.1*")
        self.assertEqual(result, "=*")



# Copyright 2008-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import functools
import logging
import signal
import sys

from _emerge.SubProcess import SubProcess
import portage
from portage import os
from portage.const import BASH_BINARY
from portage.localization import _
from portage.output import EOutput
from portage.util import writemsg_level
from portage.util._async.BuildLogger import BuildLogger
from portage.util._async.PipeLogger import PipeLogger
from portage.util._pty import _create_pty_or_pipe
from portage.util.futures import asyncio


class SpawnProcess(SubProcess):

    """
    Constructor keyword args are passed into portage.process.spawn().
    The required "args" keyword argument will be passed as the first
    spawn() argument.
    """

    _spawn_kwarg_names = (
        "env",
        "opt_name",
        "fd_pipes",
        "uid",
        "gid",
        "groups",
        "umask",
        "logfile",
        "path_lookup",
        "pre_exec",
        "close_fds",
        "cgroup",
        "unshare_ipc",
        "unshare_mount",
        "unshare_pid",
        "unshare_net",
    )

    __slots__ = (
        ("args", "log_filter_file")
        + _spawn_kwarg_names
        + (
            "_main_task",
            "_main_task_cancel",
            "_selinux_type",
        )
    )

    # Max number of attempts to kill the processes listed in cgroup.procs,
    # given that processes may fork before they can be killed.
    _CGROUP_CLEANUP_RETRY_MAX = 8

    def _start(self):

        if self.fd_pipes is None:
            self.fd_pipes = {}
        else:
            self.fd_pipes = self.fd_pipes.copy()
        fd_pipes = self.fd_pipes

        master_fd, slave_fd = self._pipe(fd_pipes)

        can_log = self._can_log(slave_fd)
        if can_log:
            log_file_path = self.logfile
        else:
            log_file_path = None

        null_input = None
        if not self.background or 0 in fd_pipes:
            # Subclasses such as AbstractEbuildProcess may have already passed
            # in a null file descriptor in fd_pipes, so use that when given.
            pass
        else:
            # TODO: Use job control functions like tcsetpgrp() to control
            # access to stdin. Until then, use /dev/null so that any
            # attempts to read from stdin will immediately return EOF
            # instead of blocking indefinitely.
            null_input = os.open("/dev/null", os.O_RDWR)
            fd_pipes[0] = null_input

        fd_pipes.setdefault(0, portage._get_stdin().fileno())
        fd_pipes.setdefault(1, sys.__stdout__.fileno())
        fd_pipes.setdefault(2, sys.__stderr__.fileno())

        # flush any pending output
        stdout_filenos = (sys.__stdout__.fileno(), sys.__stderr__.fileno())
        for fd in fd_pipes.values():
            if fd in stdout_filenos:
                sys.__stdout__.flush()
                sys.__stderr__.flush()
                break

        fd_pipes_orig = fd_pipes.copy()

        if log_file_path is not None or self.background:
            fd_pipes[1] = slave_fd
            fd_pipes[2] = slave_fd

        else:
            # Create a dummy pipe that PipeLogger uses to efficiently
            # monitor for process exit by listening for the EOF event.
            # Re-use of the allocated fd number for the key in fd_pipes
            # guarantees that the keys will not collide for similarly
            # allocated pipes which are used by callers such as
            # FileDigester and MergeProcess. See the _setup_pipes
            # docstring for more benefits of this allocation approach.
            self._dummy_pipe_fd = slave_fd
            fd_pipes[slave_fd] = slave_fd

        kwargs = {}
        for k in self._spawn_kwarg_names:
            v = getattr(self, k)
            if v is not None:
                kwargs[k] = v

        kwargs["fd_pipes"] = fd_pipes
        kwargs["returnpid"] = True
        kwargs.pop("logfile", None)

        retval = self._spawn(self.args, **kwargs)

        os.close(slave_fd)
        if null_input is not None:
            os.close(null_input)

        if isinstance(retval, int):
            # spawn failed
            self.returncode = retval
            self._async_wait()
            return

        self.pid = retval[0]

        stdout_fd = None
        if can_log and not self.background:
            stdout_fd = os.dup(fd_pipes_orig[1])

        build_logger = BuildLogger(
            env=self.env,
            log_path=log_file_path,
            log_filter_file=self.log_filter_file,
            scheduler=self.scheduler,
        )
        build_logger.start()

        pipe_logger = PipeLogger(
            background=self.background,
            scheduler=self.scheduler,
            input_fd=master_fd,
            log_file_path=build_logger.stdin,
            stdout_fd=stdout_fd,
        )

        pipe_logger.start()

        self._registered = True
        self._main_task_cancel = functools.partial(
            self._main_cancel, build_logger, pipe_logger
        )
        self._main_task = asyncio.ensure_future(
            self._main(build_logger, pipe_logger, loop=self.scheduler),
            loop=self.scheduler,
        )
        self._main_task.add_done_callback(self._main_exit)

    async def _main(self, build_logger, pipe_logger, loop=None):
        try:
            if pipe_logger.poll() is None:
                await pipe_logger.async_wait()
            if build_logger.poll() is None:
                await build_logger.async_wait()
        except asyncio.CancelledError:
            self._main_cancel(build_logger, pipe_logger)
            raise

    def _main_cancel(self, build_logger, pipe_logger):
        if pipe_logger.poll() is None:
            pipe_logger.cancel()
        if build_logger.poll() is None:
            build_logger.cancel()

    def _main_exit(self, main_task):
        self._main_task = None
        self._main_task_cancel = None
        try:
            main_task.result()
        except asyncio.CancelledError:
            self.cancel()
        self._async_waitpid()

    def _async_wait(self):
        # Allow _main_task to exit normally rather than via cancellation.
        if self._main_task is None:
            super()._async_wait()

    def _async_waitpid(self):
        # Allow _main_task to exit normally rather than via cancellation.
        if self._main_task is None:
            super()._async_waitpid()

    def _can_log(self, slave_fd):
        return True

    def _pipe(self, fd_pipes):
        """
        @type fd_pipes: dict
        @param fd_pipes: pipes from which to copy terminal size if desired.
        """
        stdout_pipe = None
        if not self.background:
            stdout_pipe = fd_pipes.get(1)
        got_pty, master_fd, slave_fd = _create_pty_or_pipe(copy_term_size=stdout_pipe)
        return (master_fd, slave_fd)

    def _spawn(self, args, **kwargs):
        spawn_func = portage.process.spawn

        if self._selinux_type is not None:
            spawn_func = portage.selinux.spawn_wrapper(spawn_func, self._selinux_type)
            # bash is an allowed entrypoint, while most binaries are not
            if args[0] != BASH_BINARY:
                args = [BASH_BINARY, "-c", 'exec "$@"', args[0]] + args

        return spawn_func(args, **kwargs)

    def _unregister(self):
        SubProcess._unregister(self)
        if self.cgroup is not None:
            self._cgroup_cleanup()
            self.cgroup = None
        if self._main_task is not None:
            self._main_task.done() or self._main_task.cancel()

    def _cancel(self):
        if self._main_task is not None:
            if not self._main_task.done():
                if self._main_task_cancel is not None:
                    self._main_task_cancel()
                    self._main_task_cancel = None
                self._main_task.cancel()
        SubProcess._cancel(self)
        self._cgroup_cleanup()

    def _cgroup_cleanup(self):
        if self.cgroup:

            def get_pids(cgroup):
                try:
                    with open(os.path.join(cgroup, "cgroup.procs")) as f:
                        return [int(p) for p in f.read().split()]
                except OSError:
                    # removed by cgroup-release-agent
                    return []

            def kill_all(pids, sig):
                for p in pids:
                    try:
                        os.kill(p, sig)
                    except OSError as e:
                        if e.errno == errno.EPERM:
                            # Reported with hardened kernel (bug #358211).
                            writemsg_level(
                                "!!! kill: (%i) - Operation not permitted\n" % (p,),
                                level=logging.ERROR,
                                noiselevel=-1,
                            )
                        elif e.errno != errno.ESRCH:
                            raise

            # step 1: kill all orphans (loop in case of new forks)
            remaining = self._CGROUP_CLEANUP_RETRY_MAX
            while remaining:
                remaining -= 1
                pids = get_pids(self.cgroup)
                if pids:
                    kill_all(pids, signal.SIGKILL)
                else:
                    break

            if pids:
                msg = []
                msg.append(
                    _("Failed to kill pid(s) in '%(cgroup)s': %(pids)s")
                    % dict(
                        cgroup=os.path.join(self.cgroup, "cgroup.procs"),
                        pids=" ".join(str(pid) for pid in pids),
                    )
                )

                self._elog("eerror", msg)

            # step 2: remove the cgroup
            try:
                os.rmdir(self.cgroup)
            except OSError:
                # it may be removed already, or busy
                # we can't do anything good about it
                pass

    def _elog(self, elog_funcname, lines):
        elog_func = getattr(EOutput(), elog_funcname)
        for line in lines:
            elog_func(line)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import datetime
import logging
import random
import re
import signal
import socket
import sys
import tempfile
import time

from _emerge.UserQuery import UserQuery

import portage
from portage import _unicode_decode
from portage import os
from portage.const import VCS_DIRS, TIMESTAMP_FORMAT, RSYNC_PACKAGE_ATOM
from portage.output import create_color_func, yellow, blue, bold
from portage.sync.getaddrinfo_validate import getaddrinfo_validate
from portage.sync.syncbase import NewBase
from portage.util import writemsg, writemsg_level, writemsg_stdout
from portage.util.futures import asyncio

good = create_color_func("GOOD")
bad = create_color_func("BAD")
warn = create_color_func("WARN")

try:
    from gemato.exceptions import GematoException
    import gemato.openpgp
    import gemato.recursiveloader
except ImportError:
    gemato = None


SERVER_OUT_OF_DATE = -1
EXCEEDED_MAX_RETRIES = -2


class RsyncSync(NewBase):
    """Rsync sync module"""

    short_desc = "Perform sync operations on rsync based repositories"

    @staticmethod
    def name():
        return "RsyncSync"

    def __init__(self):
        NewBase.__init__(self, "rsync", RSYNC_PACKAGE_ATOM)

    def update(self):
        """Internal update function which performs the transfer"""
        opts = self.options.get("emerge_config").opts
        self.usersync_uid = self.options.get("usersync_uid", None)
        enter_invalid = "--ask-enter-invalid" in opts
        quiet = "--quiet" in opts
        out = portage.output.EOutput(quiet=quiet)
        syncuri = self.repo.sync_uri
        if self.repo.module_specific_options.get(
            "sync-rsync-vcs-ignore", "false"
        ).lower() in ("true", "yes"):
            vcs_dirs = ()
        else:
            vcs_dirs = frozenset(VCS_DIRS)
            vcs_dirs = vcs_dirs.intersection(os.listdir(self.repo.location))

        for vcs_dir in vcs_dirs:
            writemsg_level(
                (
                    "!!! %s appears to be under revision "
                    + "control (contains %s).\n!!! Aborting rsync sync "
                    '(override with "sync-rsync-vcs-ignore = true" in repos.conf).\n'
                )
                % (self.repo.location, vcs_dir),
                level=logging.ERROR,
                noiselevel=-1,
            )
            return (1, False)
        self.timeout = 180

        rsync_opts = []
        if self.settings["PORTAGE_RSYNC_OPTS"] == "":
            rsync_opts = self._set_rsync_defaults()
        else:
            rsync_opts = self._validate_rsync_opts(rsync_opts, syncuri)
        self.rsync_opts = self._rsync_opts_extend(opts, rsync_opts)

        self.extra_rsync_opts = list()
        if self.repo.module_specific_options.get("sync-rsync-extra-opts"):
            self.extra_rsync_opts.extend(
                portage.util.shlex_split(
                    self.repo.module_specific_options["sync-rsync-extra-opts"]
                )
            )

        exitcode = 0
        verify_failure = False

        # Process GLEP74 verification options.
        # Default verification to 'no'; it's enabled for ::gentoo
        # via default repos.conf though.
        self.verify_metamanifest = self.repo.module_specific_options.get(
            "sync-rsync-verify-metamanifest", "no"
        ).lower() in ("yes", "true")
        # Support overriding job count.
        self.verify_jobs = self.repo.module_specific_options.get(
            "sync-rsync-verify-jobs", None
        )
        if self.verify_jobs is not None:
            try:
                self.verify_jobs = int(self.verify_jobs)
                if self.verify_jobs < 0:
                    raise ValueError(self.verify_jobs)
            except ValueError:
                writemsg_level(
                    "!!! sync-rsync-verify-jobs not a positive integer: %s\n"
                    % (self.verify_jobs,),
                    level=logging.WARNING,
                    noiselevel=-1,
                )
                self.verify_jobs = None
            else:
                if self.verify_jobs == 0:
                    # Use the apparent number of processors if gemato
                    # supports it.
                    self.verify_jobs = None
        # Support overriding max age.
        self.max_age = self.repo.module_specific_options.get(
            "sync-rsync-verify-max-age", ""
        )
        if self.max_age:
            try:
                self.max_age = int(self.max_age)
                if self.max_age < 0:
                    raise ValueError(self.max_age)
            except ValueError:
                writemsg_level(
                    "!!! sync-rsync-max-age must be a non-negative integer: %s\n"
                    % (self.max_age,),
                    level=logging.WARNING,
                    noiselevel=-1,
                )
                self.max_age = 0
        else:
            self.max_age = 0

        openpgp_env = None
        if self.verify_metamanifest and gemato is not None:
            # Use isolated environment if key is specified,
            # system environment otherwise
            openpgp_env = self._get_openpgp_env(self.repo.sync_openpgp_key_path)

        try:
            # Load and update the keyring early. If it fails, then verification
            # will not be performed and the user will have to fix it and try again,
            # so we may as well bail out before actual rsync happens.
            if openpgp_env is not None and self.repo.sync_openpgp_key_path is not None:
                try:
                    out.einfo(
                        "Using keys from {}".format(self.repo.sync_openpgp_key_path)
                    )
                    with open(self.repo.sync_openpgp_key_path, "rb") as f:
                        openpgp_env.import_key(f)
                    self._refresh_keys(openpgp_env)
                except (GematoException, asyncio.TimeoutError) as e:
                    writemsg_level(
                        "!!! Manifest verification impossible due to keyring problem:\n%s\n"
                        % (e,),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return (1, False)

            # Real local timestamp file.
            self.servertimestampfile = os.path.join(
                self.repo.location, "metadata", "timestamp.chk"
            )

            content = portage.util.grabfile(self.servertimestampfile)
            timestamp = 0
            if content:
                try:
                    timestamp = time.mktime(time.strptime(content[0], TIMESTAMP_FORMAT))
                except (OverflowError, ValueError):
                    pass
            del content

            try:
                self.rsync_initial_timeout = int(
                    self.settings.get("PORTAGE_RSYNC_INITIAL_TIMEOUT", "15")
                )
            except ValueError:
                self.rsync_initial_timeout = 15

            try:
                maxretries = int(self.settings["PORTAGE_RSYNC_RETRIES"])
            except SystemExit as e:
                raise  # Needed else can't exit
            except:
                maxretries = -1  # default number of retries

            if syncuri.startswith("file://"):
                self.proto = "file"
                dosyncuri = syncuri[7:]
                unchanged, is_synced, exitcode, updatecache_flg = self._do_rsync(
                    dosyncuri, timestamp, opts
                )
                self._process_exitcode(exitcode, dosyncuri, out, 1)
                if exitcode == 0:
                    if unchanged:
                        self.repo_storage.abort_update()
                    else:
                        self.repo_storage.commit_update()
                        self.repo_storage.garbage_collection()
                return (exitcode, updatecache_flg)

            retries = 0
            try:
                self.proto, user_name, hostname, port = re.split(
                    r"(rsync|ssh)://([^:/]+@)?(\[[:\da-fA-F]*\]|[^:/]*)(:[0-9]+)?",
                    syncuri,
                    maxsplit=4,
                )[1:5]
            except ValueError:
                writemsg_level(
                    "!!! sync-uri is invalid: %s\n" % syncuri,
                    noiselevel=-1,
                    level=logging.ERROR,
                )
                return (1, False)

            self.ssh_opts = self.settings.get("PORTAGE_SSH_OPTS")

            if port is None:
                port = ""
            if user_name is None:
                user_name = ""
            if re.match(r"^\[[:\da-fA-F]*\]$", hostname) is None:
                getaddrinfo_host = hostname
            else:
                # getaddrinfo needs the brackets stripped
                getaddrinfo_host = hostname[1:-1]
            updatecache_flg = False
            all_rsync_opts = set(self.rsync_opts)
            all_rsync_opts.update(self.extra_rsync_opts)

            family = socket.AF_UNSPEC
            if "-4" in all_rsync_opts or "--ipv4" in all_rsync_opts:
                family = socket.AF_INET
            elif socket.has_ipv6 and (
                "-6" in all_rsync_opts or "--ipv6" in all_rsync_opts
            ):
                family = socket.AF_INET6

            addrinfos = None
            uris = []

            if "RSYNC_PROXY" not in self.spawn_kwargs["env"]:
                try:
                    addrinfos = getaddrinfo_validate(
                        socket.getaddrinfo(
                            getaddrinfo_host, None, family, socket.SOCK_STREAM
                        )
                    )
                except OSError as e:
                    writemsg_level(
                        "!!! getaddrinfo failed for '%s': %s\n"
                        % (_unicode_decode(hostname), str(e)),
                        noiselevel=-1,
                        level=logging.ERROR,
                    )

            if addrinfos:

                AF_INET = socket.AF_INET
                AF_INET6 = None
                if socket.has_ipv6:
                    AF_INET6 = socket.AF_INET6

                ips_v4 = []
                ips_v6 = []

                for addrinfo in addrinfos:
                    if addrinfo[0] == AF_INET:
                        ips_v4.append("%s" % addrinfo[4][0])
                    elif AF_INET6 is not None and addrinfo[0] == AF_INET6:
                        # IPv6 addresses need to be enclosed in square brackets
                        ips_v6.append("[%s]" % addrinfo[4][0])

                random.shuffle(ips_v4)
                random.shuffle(ips_v6)

                # Give priority to the address family that
                # getaddrinfo() returned first.
                if AF_INET6 is not None and addrinfos and addrinfos[0][0] == AF_INET6:
                    ips = ips_v6 + ips_v4
                else:
                    ips = ips_v4 + ips_v6

                for ip in ips:
                    uris.append(
                        syncuri.replace(
                            "//" + user_name + hostname + port + "/",
                            "//" + user_name + ip + port + "/",
                            1,
                        )
                    )

            if not uris:
                # With some configurations we need to use the plain hostname
                # rather than try to resolve the ip addresses (bug #340817).
                uris.append(syncuri)
            elif len(uris) == 1:
                # Use the original hostname if it resolves to a single IP,
                # since DNS lookup must occur in the rsync process for
                # compatibility with things like proxychains that allocate
                # a surrogate IP which is only valid within the current
                # process.
                uris = [syncuri]

            # reverse, for use with pop()
            uris.reverse()
            uris_orig = uris[:]

            effective_maxretries = maxretries
            if effective_maxretries < 0:
                effective_maxretries = len(uris) - 1

            local_state_unchanged = True
            while 1:
                if uris:
                    dosyncuri = uris.pop()
                elif maxretries < 0 or retries > maxretries:
                    writemsg(
                        "!!! Exhausted addresses for %s\n" % _unicode_decode(hostname),
                        noiselevel=-1,
                    )
                    return (1, False)
                else:
                    uris.extend(uris_orig)
                    dosyncuri = uris.pop()

                if retries == 0:
                    if "--ask" in opts:
                        uq = UserQuery(opts)
                        if (
                            uq.query(
                                "Do you want to sync your ebuild repository "
                                + "with the mirror at\n"
                                + blue(dosyncuri)
                                + bold("?"),
                                enter_invalid,
                            )
                            == "No"
                        ):
                            print()
                            print("Quitting.")
                            print()
                            sys.exit(128 + signal.SIGINT)
                    self.logger(
                        self.xterm_titles, ">>> Starting rsync with " + dosyncuri
                    )
                    if "--quiet" not in opts:
                        print(">>> Starting rsync with " + dosyncuri + "...")
                else:
                    self.logger(
                        self.xterm_titles,
                        ">>> Starting retry %d of %d with %s"
                        % (retries, effective_maxretries, dosyncuri),
                    )
                    writemsg_stdout(
                        "\n\n>>> Starting retry %d of %d with %s\n"
                        % (retries, effective_maxretries, dosyncuri),
                        noiselevel=-1,
                    )

                if dosyncuri.startswith("ssh://"):
                    dosyncuri = dosyncuri[6:].replace("/", ":/", 1)

                unchanged, is_synced, exitcode, updatecache_flg = self._do_rsync(
                    dosyncuri, timestamp, opts
                )
                if not unchanged:
                    local_state_unchanged = False
                if is_synced:
                    break

                retries = retries + 1

                if maxretries < 0 or retries <= maxretries:
                    print(">>> Retrying...")
                else:
                    # over retries
                    # exit loop
                    exitcode = EXCEEDED_MAX_RETRIES
                    break

            self._process_exitcode(exitcode, dosyncuri, out, maxretries)

            if local_state_unchanged:
                # The quarantine download_dir is not intended to exist
                # in this case, so refer gemato to the normal repository
                # location.
                download_dir = self.repo.location
            else:
                download_dir = self.download_dir

            # if synced successfully, verify now
            if exitcode == 0 and self.verify_metamanifest:
                if gemato is None:
                    writemsg_level(
                        "!!! Unable to verify: gemato-14.5+ is required\n",
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    exitcode = 127
                else:
                    try:
                        # we always verify the Manifest signature, in case
                        # we had to deal with key revocation case
                        m = gemato.recursiveloader.ManifestRecursiveLoader(
                            os.path.join(download_dir, "Manifest"),
                            verify_openpgp=True,
                            openpgp_env=openpgp_env,
                            max_jobs=self.verify_jobs,
                        )
                        if not m.openpgp_signed:
                            raise RuntimeError(
                                "OpenPGP signature not found on Manifest"
                            )

                        ts = m.find_timestamp()
                        if ts is None:
                            raise RuntimeError("Timestamp not found in Manifest")
                        if (
                            self.max_age != 0
                            and (datetime.datetime.utcnow() - ts.ts).days > self.max_age
                        ):
                            out.quiet = False
                            out.ewarn(
                                "Manifest is over %d days old, this is suspicious!"
                                % (self.max_age,)
                            )
                            out.ewarn(
                                "You may want to try using another mirror and/or reporting this one:"
                            )
                            out.ewarn("  {}".format(dosyncuri))
                            out.ewarn("")
                            out.quiet = quiet

                        out.einfo("Manifest timestamp: {} UTC".format(ts.ts))
                        out.einfo("Valid OpenPGP signature found:")
                        out.einfo(
                            "- primary key: %s"
                            % (m.openpgp_signature.primary_key_fingerprint)
                        )
                        out.einfo("- subkey: %s" % (m.openpgp_signature.fingerprint))
                        out.einfo(
                            "- timestamp: %s UTC" % (m.openpgp_signature.timestamp)
                        )

                        # if nothing has changed, skip the actual Manifest
                        # verification
                        if not local_state_unchanged:
                            out.ebegin("Verifying {}".format(download_dir))
                            m.assert_directory_verifies()
                            out.eend(0)
                    except GematoException as e:
                        writemsg_level(
                            "!!! Manifest verification failed:\n{}\n".format(e),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        exitcode = 1
                        verify_failure = True

            if exitcode == 0 and not local_state_unchanged:
                self.repo_storage.commit_update()
                self.repo_storage.garbage_collection()

            return (exitcode, updatecache_flg)
        finally:
            # Don't delete the update if verification failed, in case
            # the cause needs to be investigated.
            if not verify_failure:
                self.repo_storage.abort_update()
            if openpgp_env is not None:
                openpgp_env.close()

    def _process_exitcode(self, exitcode, syncuri, out, maxretries):
        if exitcode == 0:
            pass
        elif exitcode == SERVER_OUT_OF_DATE:
            exitcode = 1
        elif exitcode == EXCEEDED_MAX_RETRIES:
            sys.stderr.write(">>> Exceeded PORTAGE_RSYNC_RETRIES: %s\n" % maxretries)
            exitcode = 1
        elif exitcode > 0:
            msg = []
            if exitcode == 1:
                msg.append(
                    "Rsync has reported that there is a syntax error. Please ensure"
                )
                msg.append(
                    "that sync-uri attribute for repository '%s' is proper."
                    % self.repo.name
                )
                msg.append("sync-uri: '%s'" % self.repo.sync_uri)
            elif exitcode == 11:
                msg.append("Rsync has reported that there is a File IO error. Normally")
                msg.append(
                    "this means your disk is full, but can be caused by corruption"
                )
                msg.append(
                    "on the filesystem that contains repository '%s'. Please investigate"
                    % self.repo.name
                )
                msg.append("and try again after the problem has been fixed.")
                msg.append("Location of repository: '%s'" % self.repo.location)
            elif exitcode == 20:
                msg.append("Rsync was killed before it finished.")
            else:
                msg.append(
                    "Rsync has not successfully finished. It is recommended that you keep"
                )
                msg.append(
                    "trying or that you use the 'emerge-webrsync' option if you are unable"
                )
                msg.append(
                    "to use rsync due to firewall or other restrictions. This should be a"
                )
                msg.append(
                    "temporary problem unless complications exist with your network"
                )
                msg.append("(and possibly your system's filesystem) configuration.")
            for line in msg:
                out.eerror(line)

    def new(self, **kwargs):
        if kwargs:
            self._kwargs(kwargs)
        try:
            if not os.path.exists(self.repo.location):
                os.makedirs(self.repo.location)
                self.logger(
                    self.self.xterm_titles,
                    "Created New Directory %s " % self.repo.location,
                )
        except OSError:
            return (1, False)
        return self.update()

    def retrieve_head(self, **kwargs):
        """Get information about the head commit"""
        if kwargs:
            self._kwargs(kwargs)
        last_sync = portage.grabfile(
            os.path.join(self.repo.location, "metadata", "timestamp.commit")
        )
        ret = (1, False)
        if last_sync:
            try:
                ret = (os.EX_OK, last_sync[0].split()[0])
            except IndexError:
                pass
        return ret

    def _set_rsync_defaults(self):
        portage.writemsg(
            "PORTAGE_RSYNC_OPTS empty or unset, using hardcoded defaults\n"
        )
        rsync_opts = [
            "--recursive",  # Recurse directories
            "--links",  # Consider symlinks
            "--safe-links",  # Ignore links outside of tree
            "--perms",  # Preserve permissions
            "--times",  # Preserive mod times
            "--omit-dir-times",
            "--compress",  # Compress the data transmitted
            "--force",  # Force deletion on non-empty dirs
            "--whole-file",  # Don't do block transfers, only entire files
            "--delete",  # Delete files that aren't in the master tree
            "--stats",  # Show final statistics about what was transfered
            "--human-readable",
            "--timeout=" + str(self.timeout),  # IO timeout if not done in X seconds
            "--exclude=/distfiles",  # Exclude distfiles from consideration
            "--exclude=/local",  # Exclude local     from consideration
            "--exclude=/packages",  # Exclude packages  from consideration
        ]
        return rsync_opts

    def _validate_rsync_opts(self, rsync_opts, syncuri):
        # The below validation is not needed when using the above hardcoded
        # defaults.

        portage.writemsg("Using PORTAGE_RSYNC_OPTS instead of hardcoded defaults\n", 1)
        rsync_opts.extend(
            portage.util.shlex_split(self.settings.get("PORTAGE_RSYNC_OPTS", ""))
        )
        for opt in ("--recursive", "--times"):
            if opt not in rsync_opts:
                portage.writemsg(
                    yellow("WARNING:")
                    + " adding required option "
                    + "%s not included in PORTAGE_RSYNC_OPTS\n" % opt
                )
                rsync_opts.append(opt)

        for exclude in ("distfiles", "local", "packages"):
            opt = "--exclude=/%s" % exclude
            if opt not in rsync_opts:
                portage.writemsg(
                    yellow("WARNING:")
                    + " adding required option %s not included in " % opt
                    + "PORTAGE_RSYNC_OPTS (can be overridden with --exclude='!')\n"
                )
                rsync_opts.append(opt)

        if syncuri.rstrip("/").endswith(".gentoo.org/gentoo-portage"):

            def rsync_opt_startswith(opt_prefix):
                for x in rsync_opts:
                    if x.startswith(opt_prefix):
                        return (1, False)
                return (0, False)

            if not rsync_opt_startswith("--timeout="):
                rsync_opts.append("--timeout=%d" % self.timeout)

            for opt in ("--compress", "--whole-file"):
                if opt not in rsync_opts:
                    portage.writemsg(
                        yellow("WARNING:")
                        + " adding required option "
                        + "%s not included in PORTAGE_RSYNC_OPTS\n" % opt
                    )
                    rsync_opts.append(opt)
        return rsync_opts

    @staticmethod
    def _rsync_opts_extend(opts, rsync_opts):
        if "--quiet" in opts:
            rsync_opts.append("--quiet")  # Shut up a lot
        else:
            rsync_opts.append("--verbose")  # Print filelist

        if "--verbose" in opts:
            rsync_opts.append("--progress")  # Progress meter for each file

        if "--debug" in opts:
            rsync_opts.append("--checksum")  # Force checksum on all files
        return rsync_opts

    def _do_rsync(self, syncuri, timestamp, opts):
        updatecache_flg = False
        is_synced = False
        if timestamp != 0 and "--quiet" not in opts:
            print(">>> Checking server timestamp ...")

        rsynccommand = [self.bin_command] + self.rsync_opts + self.extra_rsync_opts

        if self.proto == "ssh" and self.ssh_opts:
            rsynccommand.append("--rsh=ssh " + self.ssh_opts)

        if "--debug" in opts:
            print(rsynccommand)

        local_state_unchanged = False
        exitcode = os.EX_OK
        servertimestamp = 0
        # Even if there's no timestamp available locally, fetch the
        # timestamp anyway as an initial probe to verify that the server is
        # responsive.  This protects us from hanging indefinitely on a
        # connection attempt to an unresponsive server which rsync's
        # --timeout option does not prevent.

        # if True:
        # Temporary file for remote server timestamp comparison.
        # NOTE: If FEATURES=usersync is enabled then the tempfile
        # needs to be in a directory that's readable by the usersync
        # user. We assume that ${PORTAGE_TMPDIR}/portage will satisfy this
        # requirement, since that's not necessarily true for the
        # default directory used by the tempfile module.
        if self.usersync_uid is not None:
            tmpdir = os.path.join(self.settings["PORTAGE_TMPDIR"], "portage")
            ensure_dirs_kwargs = {}
            if portage.secpass >= 1:
                ensure_dirs_kwargs["gid"] = portage.portage_gid
                ensure_dirs_kwargs["mode"] = 0o70
                ensure_dirs_kwargs["mask"] = 0
            portage.util.ensure_dirs(tmpdir, **ensure_dirs_kwargs)
        else:
            # use default dir from tempfile module
            tmpdir = None
        fd, tmpservertimestampfile = tempfile.mkstemp(dir=tmpdir)
        os.close(fd)
        if self.usersync_uid is not None:
            portage.util.apply_permissions(
                tmpservertimestampfile, uid=self.usersync_uid
            )
        command = rsynccommand[:]
        command.append("--inplace")
        command.append(syncuri.rstrip("/") + "/metadata/timestamp.chk")
        command.append(tmpservertimestampfile)
        content = None
        pids = []
        try:
            # Timeout here in case the server is unresponsive.  The
            # --timeout rsync option doesn't apply to the initial
            # connection attempt.
            try:
                if self.rsync_initial_timeout:
                    portage.exception.AlarmSignal.register(self.rsync_initial_timeout)

                pids.extend(
                    portage.process.spawn(command, returnpid=True, **self.spawn_kwargs)
                )
                exitcode = os.waitpid(pids[0], 0)[1]
                if self.usersync_uid is not None:
                    portage.util.apply_permissions(
                        tmpservertimestampfile, uid=os.getuid()
                    )
                content = portage.grabfile(tmpservertimestampfile)
            finally:
                if self.rsync_initial_timeout:
                    portage.exception.AlarmSignal.unregister()
                try:
                    os.unlink(tmpservertimestampfile)
                except OSError:
                    pass
        except portage.exception.AlarmSignal:
            # timed out
            print("timed out")
            # With waitpid and WNOHANG, only check the
            # first element of the tuple since the second
            # element may vary (bug #337465).
            if pids and os.waitpid(pids[0], os.WNOHANG)[0] == 0:
                os.kill(pids[0], signal.SIGTERM)
                os.waitpid(pids[0], 0)
            # This is the same code rsync uses for timeout.
            exitcode = 30
        else:
            if exitcode != os.EX_OK:
                if exitcode & 0xFF:
                    exitcode = (exitcode & 0xFF) << 8
                else:
                    exitcode = exitcode >> 8

        if content:
            try:
                servertimestamp = time.mktime(
                    time.strptime(content[0], TIMESTAMP_FORMAT)
                )
            except (OverflowError, ValueError):
                pass
        del command, pids, content

        if exitcode == os.EX_OK:
            if (servertimestamp != 0) and (servertimestamp == timestamp):
                local_state_unchanged = True
                is_synced = True
                self.logger(
                    self.xterm_titles, ">>> Cancelling sync -- Already current."
                )
                print()
                print(">>>")
                print(
                    ">>> Timestamps on the server and in the local repository are the same."
                )
                print(
                    ">>> Cancelling all further sync action. You are already up to date."
                )
                print(">>>")
                print(
                    ">>> In order to force sync, remove '%s'."
                    % self.servertimestampfile
                )
                print(">>>")
                print()
            elif (servertimestamp != 0) and (servertimestamp < timestamp):
                self.logger(self.xterm_titles, ">>> Server out of date: %s" % syncuri)
                print()
                print(">>>")
                print(">>> SERVER OUT OF DATE: %s" % syncuri)
                print(">>>")
                print(
                    ">>> In order to force sync, remove '%s'."
                    % self.servertimestampfile
                )
                print(">>>")
                print()
                exitcode = SERVER_OUT_OF_DATE
            elif (servertimestamp == 0) or (servertimestamp > timestamp):
                # actual sync
                command = rsynccommand[:]

                submodule_paths = self._get_submodule_paths()
                if submodule_paths:
                    # The only way to select multiple directories to
                    # sync, without calling rsync multiple times, is
                    # to use --relative.
                    command.append("--relative")
                    for path in submodule_paths:
                        # /./ is special syntax supported with the
                        # rsync --relative option.
                        command.append(syncuri + "/./" + path)
                else:
                    command.append(syncuri + "/")

                command.append(self.download_dir)

                exitcode = None
                try:
                    exitcode = portage.process.spawn(command, **self.spawn_kwargs)
                finally:
                    if exitcode is None:
                        # interrupted
                        exitcode = 128 + signal.SIGINT

                    #   0	Success
                    #   1	Syntax or usage error
                    #   2	Protocol incompatibility
                    #   5	Error starting client-server protocol
                    #  35	Timeout waiting for daemon connection
                    if exitcode not in (0, 1, 2, 5, 35):
                        # If the exit code is not among those listed above,
                        # then we may have a partial/inconsistent sync
                        # state, so our previously read timestamp as well
                        # as the corresponding file can no longer be
                        # trusted.
                        timestamp = 0
                        try:
                            os.unlink(self.servertimestampfile)
                        except OSError:
                            pass
                    else:
                        updatecache_flg = True

                if exitcode in [0, 1, 3, 4, 11, 14, 20, 21]:
                    is_synced = True
        elif exitcode in [1, 3, 4, 11, 14, 20, 21]:
            is_synced = True
        else:
            # Code 2 indicates protocol incompatibility, which is expected
            # for servers with protocol < 29 that don't support
            # --prune-empty-directories.  Retry for a server that supports
            # at least rsync protocol version 29 (>=rsync-2.6.4).
            pass

        return local_state_unchanged, is_synced, exitcode, updatecache_flg



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import stat

from portage import os
from portage.exception import PermissionDenied


def exists_raise_eaccess(path):
    try:
        os.stat(path)
    except OSError as e:
        if e.errno == PermissionDenied.errno:
            raise PermissionDenied("stat('%s')" % path)
        return False
    else:
        return True


def isdir_raise_eaccess(path):
    try:
        st = os.stat(path)
    except OSError as e:
        if e.errno == PermissionDenied.errno:
            raise PermissionDenied("stat('%s')" % path)
        return False
    else:
        return stat.S_ISDIR(st.st_mode)



# test_dep_getusedeps.py -- Portage Unit Testing Functionality
# Copyright 2007-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import dep_getusedeps

from portage.tests import test_cps, test_slots, test_versions, test_usedeps


class DepGetUseDeps(TestCase):
    """A simple testcase for dep_getusedeps"""

    def testDepGetUseDeps(self):

        for mycpv in test_cps:
            for version in test_versions:
                for slot in test_slots:
                    for use in test_usedeps:
                        cpv = mycpv[:]
                        if version:
                            cpv += version
                        if slot:
                            cpv += ":" + slot
                        if isinstance(use, tuple):
                            cpv += "[{}]".format(",".join(use))
                            self.assertEqual(dep_getusedeps(cpv), use)
                        else:
                            if len(use):
                                self.assertEqual(
                                    dep_getusedeps(cpv + "[" + use + "]"), (use,)
                                )
                            else:
                                self.assertEqual(
                                    dep_getusedeps(cpv + "[" + use + "]"), ()
                                )



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictUpdateTestCase(TestCase):
    def testSlotConflictUpdate(self):

        ebuilds = {
            "app-text/podofo-0.9.2": {"EAPI": "5", "RDEPEND": "dev-util/boost-build"},
            "dev-cpp/libcmis-0.3.1": {"EAPI": "5", "RDEPEND": "dev-libs/boost:="},
            "dev-libs/boost-1.53.0": {
                "EAPI": "5",
                "SLOT": "0/1.53",
                "RDEPEND": "=dev-util/boost-build-1.53.0",
            },
            "dev-libs/boost-1.52.0": {
                "EAPI": "5",
                "SLOT": "0/1.52",
                "RDEPEND": "=dev-util/boost-build-1.52.0",
            },
            "dev-util/boost-build-1.53.0": {"EAPI": "5", "SLOT": "0"},
            "dev-util/boost-build-1.52.0": {"EAPI": "5", "SLOT": "0"},
        }

        installed = {
            "app-text/podofo-0.9.2": {"EAPI": "5", "RDEPEND": "dev-util/boost-build"},
            "dev-cpp/libcmis-0.3.1": {"EAPI": "5", "RDEPEND": "dev-libs/boost:0/1.52="},
            "dev-util/boost-build-1.52.0": {"EAPI": "5", "SLOT": "0"},
            "dev-libs/boost-1.52.0": {
                "EAPI": "5",
                "SLOT": "0/1.52",
                "RDEPEND": "=dev-util/boost-build-1.52.0",
            },
        }

        world = ["dev-cpp/libcmis", "dev-libs/boost", "app-text/podofo"]

        test_cases = (
            # In order to avoid a missed update, first mask lower
            # versions that conflict with higher versions. Note that
            # this behavior makes SlotConflictMaskUpdateTestCase
            # fail.
            ResolverPlaygroundTestCase(
                ["@world"],
                all_permutations=True,
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "dev-util/boost-build-1.53.0",
                    "dev-libs/boost-1.53.0",
                    "dev-cpp/libcmis-0.3.1",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections
import errno
import itertools
import logging
import subprocess

import portage
from portage import _encodings
from portage import _os_merge
from portage import _unicode_decode
from portage import _unicode_encode
from portage.cache.mappings import slot_dict_class
from portage.const import EPREFIX
from portage.dep.soname.multilib_category import compute_multilib_category
from portage.dep.soname.SonameAtom import SonameAtom
from portage.exception import CommandNotFound, InvalidData
from portage.localization import _
from portage.util import getlibpaths
from portage.util import grabfile
from portage.util import normalize_path
from portage.util import varexpand
from portage.util import writemsg_level
from portage.util._dyn_libs.NeededEntry import NeededEntry
from portage.util.elf.header import ELFHeader


# Map ELF e_machine values from NEEDED.ELF.2 to approximate multilib
# categories. This approximation will produce incorrect results on x32
# and mips systems, but the result is not worse than using the raw
# e_machine value which was used by earlier versions of portage.
_approx_multilib_categories = {
    "386": "x86_32",
    "68K": "m68k_32",
    "AARCH64": "arm_64",
    "ALPHA": "alpha_64",
    "ARM": "arm_32",
    "IA_64": "ia64_64",
    "MIPS": "mips_o32",
    "PARISC": "hppa_64",
    "PPC": "ppc_32",
    "PPC64": "ppc_64",
    "S390": "s390_64",
    "SH": "sh_32",
    "SPARC": "sparc_32",
    "SPARC32PLUS": "sparc_32",
    "SPARCV9": "sparc_64",
    "X86_64": "x86_64",
}


class LinkageMapELF:

    """Models dynamic linker dependencies."""

    _needed_aux_key = "NEEDED.ELF.2"
    _soname_map_class = slot_dict_class(("consumers", "providers"), prefix="")

    class _obj_properties_class:

        __slots__ = (
            "arch",
            "needed",
            "runpaths",
            "soname",
            "alt_paths",
            "owner",
        )

        def __init__(self, arch, needed, runpaths, soname, alt_paths, owner):
            self.arch = arch
            self.needed = needed
            self.runpaths = runpaths
            self.soname = soname
            self.alt_paths = alt_paths
            self.owner = owner

    def __init__(self, vardbapi):
        self._dbapi = vardbapi
        self._root = self._dbapi.settings["ROOT"]
        self._libs = {}
        self._obj_properties = {}
        self._obj_key_cache = {}
        self._defpath = set()
        self._path_key_cache = {}

    def _clear_cache(self):
        self._libs.clear()
        self._obj_properties.clear()
        self._obj_key_cache.clear()
        self._defpath.clear()
        self._path_key_cache.clear()

    def _path_key(self, path):
        key = self._path_key_cache.get(path)
        if key is None:
            key = self._ObjectKey(path, self._root)
            self._path_key_cache[path] = key
        return key

    def _obj_key(self, path):
        key = self._obj_key_cache.get(path)
        if key is None:
            key = self._ObjectKey(path, self._root)
            self._obj_key_cache[path] = key
        return key

    class _ObjectKey:

        """Helper class used as _obj_properties keys for objects."""

        __slots__ = ("_key",)

        def __init__(self, obj, root):
            """
            This takes a path to an object.

            @param object: path to a file
            @type object: string (example: '/usr/bin/bar')

            """
            self._key = self._generate_object_key(obj, root)

        def __hash__(self):
            return hash(self._key)

        def __eq__(self, other):
            return self._key == other._key

        def _generate_object_key(self, obj, root):
            """
            Generate object key for a given object.

            @param object: path to a file
            @type object: string (example: '/usr/bin/bar')
            @rtype: 2-tuple of types (long, int) if object exists. string if
                    object does not exist.
            @return:
                    1. 2-tuple of object's inode and device from a stat call, if object
                            exists.
                    2. realpath of object if object does not exist.

            """

            os = _os_merge

            try:
                _unicode_encode(obj, encoding=_encodings["merge"], errors="strict")
            except UnicodeEncodeError:
                # The package appears to have been merged with a
                # different value of sys.getfilesystemencoding(),
                # so fall back to utf_8 if appropriate.
                try:
                    _unicode_encode(obj, encoding=_encodings["fs"], errors="strict")
                except UnicodeEncodeError:
                    pass
                else:
                    os = portage.os

            abs_path = os.path.join(root, obj.lstrip(os.sep))
            try:
                object_stat = os.stat(abs_path)
            except OSError:
                # Use the realpath as the key if the file does not exists on the
                # filesystem.
                return os.path.realpath(abs_path)
            # Return a tuple of the device and inode.
            return (object_stat.st_dev, object_stat.st_ino)

        def file_exists(self):
            """
            Determine if the file for this key exists on the filesystem.

            @rtype: Boolean
            @return:
                    1. True if the file exists.
                    2. False if the file does not exist or is a broken symlink.

            """
            return isinstance(self._key, tuple)

    class _LibGraphNode(_ObjectKey):
        __slots__ = ("alt_paths",)

        def __init__(self, key):
            """
            Create a _LibGraphNode from an existing _ObjectKey.
            This re-uses the _key attribute in order to avoid repeating
            any previous stat calls, which helps to avoid potential race
            conditions due to inconsistent stat results when the
            file system is being modified concurrently.
            """
            self._key = key._key
            self.alt_paths = set()

        def __str__(self):
            return str(sorted(self.alt_paths))

    def rebuild(self, exclude_pkgs=None, include_file=None, preserve_paths=None):
        """
        Raises CommandNotFound if there are preserved libs
        and the scanelf binary is not available.

        @param exclude_pkgs: A set of packages that should be excluded from
                the LinkageMap, since they are being unmerged and their NEEDED
                entries are therefore irrelevant and would only serve to corrupt
                the LinkageMap.
        @type exclude_pkgs: set
        @param include_file: The path of a file containing NEEDED entries for
                a package which does not exist in the vardbapi yet because it is
                currently being merged.
        @type include_file: String
        @param preserve_paths: Libraries preserved by a package instance that
                is currently being merged. They need to be explicitly passed to the
                LinkageMap, since they are not registered in the
                PreservedLibsRegistry yet.
        @type preserve_paths: set
        """

        os = _os_merge
        root = self._root
        root_len = len(root) - 1
        self._clear_cache()
        self._defpath.update(
            getlibpaths(self._dbapi.settings["EROOT"], env=self._dbapi.settings)
        )
        libs = self._libs
        obj_properties = self._obj_properties

        lines = []

        # Data from include_file is processed first so that it
        # overrides any data from previously installed files.
        if include_file is not None:
            for line in grabfile(include_file):
                lines.append((None, include_file, line))

        aux_keys = [self._needed_aux_key]
        can_lock = os.access(os.path.dirname(self._dbapi._dbroot), os.W_OK)
        if can_lock:
            self._dbapi.lock()
        try:
            for cpv in self._dbapi.cpv_all():
                if exclude_pkgs is not None and cpv in exclude_pkgs:
                    continue
                needed_file = self._dbapi.getpath(cpv, filename=self._needed_aux_key)
                for line in self._dbapi.aux_get(cpv, aux_keys)[0].splitlines():
                    lines.append((cpv, needed_file, line))
        finally:
            if can_lock:
                self._dbapi.unlock()

        # have to call scanelf for preserved libs here as they aren't
        # registered in NEEDED.ELF.2 files
        plibs = {}
        if preserve_paths is not None:
            plibs.update((x, None) for x in preserve_paths)
        if self._dbapi._plib_registry and self._dbapi._plib_registry.hasEntries():
            for cpv, items in self._dbapi._plib_registry.getPreservedLibs().items():
                if exclude_pkgs is not None and cpv in exclude_pkgs:
                    # These preserved libs will either be unmerged,
                    # rendering them irrelevant, or they will be
                    # preserved in the replacement package and are
                    # already represented via the preserve_paths
                    # parameter.
                    continue
                plibs.update((x, cpv) for x in items)
        if plibs:
            # We don't use scanelf -q, since that would omit libraries like
            # musl's /usr/lib/libc.so which do not have any DT_NEEDED or
            # DT_SONAME settings.
            args = [
                os.path.join(EPREFIX or "/", "usr/bin/scanelf"),
                "-BF",
                "%a;%F;%S;%r;%n",
            ]
            args.extend(os.path.join(root, x.lstrip("." + os.sep)) for x in plibs)
            try:
                proc = subprocess.Popen(args, stdout=subprocess.PIPE)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                raise CommandNotFound(args[0])
            else:
                for l in proc.stdout:
                    try:
                        l = _unicode_decode(
                            l, encoding=_encodings["content"], errors="strict"
                        )
                    except UnicodeDecodeError:
                        l = _unicode_decode(
                            l, encoding=_encodings["content"], errors="replace"
                        )
                        writemsg_level(
                            _(
                                "\nError decoding characters "
                                "returned from scanelf: %s\n\n"
                            )
                            % (l,),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                    l = l[3:].rstrip("\n")
                    if not l:
                        continue
                    try:
                        entry = NeededEntry.parse("scanelf", l)
                    except InvalidData as e:
                        writemsg_level(
                            "\n{}\n\n".format(e), level=logging.ERROR, noiselevel=-1
                        )
                        continue
                    try:
                        with open(
                            _unicode_encode(
                                entry.filename,
                                encoding=_encodings["fs"],
                                errors="strict",
                            ),
                            "rb",
                        ) as f:
                            elf_header = ELFHeader.read(f)
                    except OSError as e:
                        if e.errno != errno.ENOENT:
                            raise
                        # File removed concurrently.
                        continue

                    # Infer implicit soname from basename (bug 715162).
                    if not entry.soname:
                        try:
                            proc = subprocess.Popen(
                                [
                                    b"file",
                                    _unicode_encode(
                                        entry.filename,
                                        encoding=_encodings["fs"],
                                        errors="strict",
                                    ),
                                ],
                                stdout=subprocess.PIPE,
                            )
                            out, err = proc.communicate()
                            proc.wait()
                        except OSError:
                            pass
                        else:
                            if b"SB shared object" in out:
                                entry.soname = os.path.basename(entry.filename)

                    entry.multilib_category = compute_multilib_category(elf_header)
                    entry.filename = entry.filename[root_len:]
                    owner = plibs.pop(entry.filename, None)
                    lines.append((owner, "scanelf", str(entry)))
                proc.wait()
                proc.stdout.close()

        if plibs:
            # Preserved libraries that did not appear in the scanelf output.
            # This is known to happen with statically linked libraries.
            # Generate dummy lines for these, so we can assume that every
            # preserved library has an entry in self._obj_properties. This
            # is important in order to prevent findConsumers from raising
            # an unwanted KeyError.
            for x, cpv in plibs.items():
                lines.append((cpv, "plibs", ";".join(["", x, "", "", ""])))

        # Share identical frozenset instances when available,
        # in order to conserve memory.
        frozensets = {}
        owner_entries = collections.defaultdict(list)

        while True:
            try:
                owner, location, l = lines.pop()
            except IndexError:
                break
            l = l.rstrip("\n")
            if not l:
                continue
            if "\0" in l:
                # os.stat() will raise "TypeError: must be encoded string
                # without NULL bytes, not str" in this case.
                writemsg_level(
                    _("\nLine contains null byte(s) " "in %s: %s\n\n") % (location, l),
                    level=logging.ERROR,
                    noiselevel=-1,
                )
                continue
            try:
                entry = NeededEntry.parse(location, l)
            except InvalidData as e:
                writemsg_level("\n{}\n\n".format(e), level=logging.ERROR, noiselevel=-1)
                continue

            # If NEEDED.ELF.2 contains the new multilib category field,
            # then use that for categorization. Otherwise, if a mapping
            # exists, map e_machine (entry.arch) to an approximate
            # multilib category. If all else fails, use e_machine, just
            # as older versions of portage did.
            if entry.multilib_category is None:
                entry.multilib_category = _approx_multilib_categories.get(
                    entry.arch, entry.arch
                )

            entry.filename = normalize_path(entry.filename)
            expand = {"ORIGIN": os.path.dirname(entry.filename)}
            entry.runpaths = frozenset(
                normalize_path(
                    varexpand(x, expand, error_leader=lambda: "%s: " % location)
                )
                for x in entry.runpaths
            )
            entry.runpaths = frozensets.setdefault(entry.runpaths, entry.runpaths)
            owner_entries[owner].append(entry)

        # In order to account for internal library resolution which a package
        # may implement (useful at least for handling of bundled libraries),
        # generate implicit runpath entries for any needed sonames which are
        # provided by the same owner package.
        for owner, entries in owner_entries.items():
            if owner is None:
                continue

            providers = {}
            for entry in entries:
                if entry.soname:
                    providers[SonameAtom(entry.multilib_category, entry.soname)] = entry

            for entry in entries:
                implicit_runpaths = []
                for soname in entry.needed:
                    soname_atom = SonameAtom(entry.multilib_category, soname)
                    provider = providers.get(soname_atom)
                    if provider is None:
                        continue
                    provider_dir = os.path.dirname(provider.filename)
                    if provider_dir not in entry.runpaths:
                        implicit_runpaths.append(provider_dir)

                if implicit_runpaths:
                    entry.runpaths = frozenset(
                        itertools.chain(entry.runpaths, implicit_runpaths)
                    )
                    entry.runpaths = frozensets.setdefault(
                        entry.runpaths, entry.runpaths
                    )

        for owner, entry in (
            (owner, entry)
            for (owner, entries) in owner_entries.items()
            for entry in entries
        ):
            arch = entry.multilib_category
            obj = entry.filename
            soname = entry.soname
            path = entry.runpaths
            needed = frozenset(entry.needed)

            needed = frozensets.setdefault(needed, needed)

            obj_key = self._obj_key(obj)
            indexed = True
            myprops = obj_properties.get(obj_key)
            if myprops is None:
                indexed = False
                myprops = self._obj_properties_class(
                    arch, needed, path, soname, [], owner
                )
                obj_properties[obj_key] = myprops
            # All object paths are added into the obj_properties tuple.
            myprops.alt_paths.append(obj)

            # Don't index the same file more that once since only one
            # set of data can be correct and therefore mixing data
            # may corrupt the index (include_file overrides previously
            # installed).
            if indexed:
                continue

            arch_map = libs.get(arch)
            if arch_map is None:
                arch_map = {}
                libs[arch] = arch_map
            if soname:
                soname_map = arch_map.get(soname)
                if soname_map is None:
                    soname_map = self._soname_map_class(providers=[], consumers=[])
                    arch_map[soname] = soname_map
                soname_map.providers.append(obj_key)
            for needed_soname in needed:
                soname_map = arch_map.get(needed_soname)
                if soname_map is None:
                    soname_map = self._soname_map_class(providers=[], consumers=[])
                    arch_map[needed_soname] = soname_map
                soname_map.consumers.append(obj_key)

        for arch, sonames in libs.items():
            for soname_node in sonames.values():
                soname_node.providers = tuple(set(soname_node.providers))
                soname_node.consumers = tuple(set(soname_node.consumers))

    def listBrokenBinaries(self, debug=False):
        """
        Find binaries and their needed sonames, which have no providers.

        @param debug: Boolean to enable debug output
        @type debug: Boolean
        @rtype: dict (example: {'/usr/bin/foo': set(['libbar.so'])})
        @return: The return value is an object -> set-of-sonames mapping, where
                object is a broken binary and the set consists of sonames needed by
                object that have no corresponding libraries to fulfill the dependency.

        """

        os = _os_merge

        class _LibraryCache:

            """
            Caches properties associated with paths.

            The purpose of this class is to prevent multiple instances of
            _ObjectKey for the same paths.

            """

            def __init__(cache_self):
                cache_self.cache = {}

            def get(cache_self, obj):
                """
                Caches and returns properties associated with an object.

                @param obj: absolute path (can be symlink)
                @type obj: string (example: '/usr/lib/libfoo.so')
                @rtype: 4-tuple with types
                        (string or None, string or None, 2-tuple, Boolean)
                @return: 4-tuple with the following components:
                        1. arch as a string or None if it does not exist,
                        2. soname as a string or None if it does not exist,
                        3. obj_key as 2-tuple,
                        4. Boolean representing whether the object exists.
                        (example: ('libfoo.so.1', (123L, 456L), True))

                """
                if obj in cache_self.cache:
                    return cache_self.cache[obj]

                obj_key = self._obj_key(obj)
                # Check that the library exists on the filesystem.
                if obj_key.file_exists():
                    # Get the arch and soname from LinkageMap._obj_properties if
                    # it exists. Otherwise, None.
                    obj_props = self._obj_properties.get(obj_key)
                    if obj_props is None:
                        arch = None
                        soname = None
                    else:
                        arch = obj_props.arch
                        soname = obj_props.soname
                    return cache_self.cache.setdefault(
                        obj, (arch, soname, obj_key, True)
                    )
                return cache_self.cache.setdefault(obj, (None, None, obj_key, False))

        rValue = {}
        cache = _LibraryCache()
        providers = self.listProviders()

        # Iterate over all obj_keys and their providers.
        for obj_key, sonames in providers.items():
            obj_props = self._obj_properties[obj_key]
            arch = obj_props.arch
            path = obj_props.runpaths
            objs = obj_props.alt_paths
            path = path.union(self._defpath)
            # Iterate over each needed soname and the set of library paths that
            # fulfill the soname to determine if the dependency is broken.
            for soname, libraries in sonames.items():
                # validLibraries is used to store libraries, which satisfy soname,
                # so if no valid libraries are found, the soname is not satisfied
                # for obj_key.  If unsatisfied, objects associated with obj_key
                # must be emerged.
                validLibraries = set()
                # It could be the case that the library to satisfy the soname is
                # not in the obj's runpath, but a symlink to the library is (eg
                # libnvidia-tls.so.1 in nvidia-drivers).  Also, since LinkageMap
                # does not catalog symlinks, broken or missing symlinks may go
                # unnoticed.  As a result of these cases, check that a file with
                # the same name as the soname exists in obj's runpath.
                # XXX If we catalog symlinks in LinkageMap, this could be improved.
                for directory in path:
                    cachedArch, cachedSoname, cachedKey, cachedExists = cache.get(
                        os.path.join(directory, soname)
                    )
                    # Check that this library provides the needed soname.  Doing
                    # this, however, will cause consumers of libraries missing
                    # sonames to be unnecessarily emerged. (eg libmix.so)
                    if cachedSoname == soname and cachedArch == arch:
                        validLibraries.add(cachedKey)
                        if debug and cachedKey not in set(
                            map(self._obj_key_cache.get, libraries)
                        ):
                            # XXX This is most often due to soname symlinks not in
                            # a library's directory.  We could catalog symlinks in
                            # LinkageMap to avoid checking for this edge case here.
                            writemsg_level(
                                _("Found provider outside of findProviders:")
                                + (
                                    " %s -> %s %s\n"
                                    % (
                                        os.path.join(directory, soname),
                                        self._obj_properties[cachedKey].alt_paths,
                                        libraries,
                                    )
                                ),
                                level=logging.DEBUG,
                                noiselevel=-1,
                            )
                        # A valid library has been found, so there is no need to
                        # continue.
                        break
                    if (
                        debug
                        and cachedArch == arch
                        and cachedKey in self._obj_properties
                    ):
                        writemsg_level(
                            (
                                _(
                                    "Broken symlink or missing/bad soname: "
                                    + "%(dir_soname)s -> %(cachedKey)s "
                                    + "with soname %(cachedSoname)s but expecting %(soname)s"
                                )
                                % {
                                    "dir_soname": os.path.join(directory, soname),
                                    "cachedKey": self._obj_properties[cachedKey],
                                    "cachedSoname": cachedSoname,
                                    "soname": soname,
                                }
                            )
                            + "\n",
                            level=logging.DEBUG,
                            noiselevel=-1,
                        )
                # This conditional checks if there are no libraries to satisfy the
                # soname (empty set).
                if not validLibraries:
                    for obj in objs:
                        rValue.setdefault(obj, set()).add(soname)
                    # If no valid libraries have been found by this point, then
                    # there are no files named with the soname within obj's runpath,
                    # but if there are libraries (from the providers mapping), it is
                    # likely that soname symlinks or the actual libraries are
                    # missing or broken.  Thus those libraries are added to rValue
                    # in order to emerge corrupt library packages.
                    for lib in libraries:
                        rValue.setdefault(lib, set()).add(soname)
                        if debug:
                            if not os.path.isfile(lib):
                                writemsg_level(
                                    _("Missing library:") + " {}\n".format(lib),
                                    level=logging.DEBUG,
                                    noiselevel=-1,
                                )
                            else:
                                writemsg_level(
                                    _("Possibly missing symlink:")
                                    + "%s\n"
                                    % (os.path.join(os.path.dirname(lib), soname)),
                                    level=logging.DEBUG,
                                    noiselevel=-1,
                                )
        return rValue

    def listProviders(self):
        """
        Find the providers for all object keys in LinkageMap.

        @rtype: dict (example:
                {(123L, 456L): {'libbar.so': set(['/lib/libbar.so.1.5'])}})
        @return: The return value is an object key -> providers mapping, where
                providers is a mapping of soname -> set-of-library-paths returned
                from the findProviders method.

        """
        rValue = {}
        if not self._libs:
            self.rebuild()
        # Iterate over all object keys within LinkageMap.
        for obj_key in self._obj_properties:
            rValue.setdefault(obj_key, self.findProviders(obj_key))
        return rValue

    def isMasterLink(self, obj):
        """
        Determine whether an object is a "master" symlink, which means
        that its basename is the same as the beginning part of the
        soname and it lacks the soname's version component.

        Examples:

        soname                 | master symlink name
        --------------------------------------------
        libarchive.so.2.8.4    | libarchive.so
        libproc-3.2.8.so       | libproc.so

        @param obj: absolute path to an object
        @type obj: string (example: '/usr/bin/foo')
        @rtype: Boolean
        @return:
                1. True if obj is a master link
                2. False if obj is not a master link

        """
        os = _os_merge
        obj_key = self._obj_key(obj)
        if obj_key not in self._obj_properties:
            raise KeyError("{} ({}) not in object list".format(obj_key, obj))
        basename = os.path.basename(obj)
        soname = self._obj_properties[obj_key].soname
        return (
            len(basename) < len(soname)
            and basename.endswith(".so")
            and soname.startswith(basename[:-3])
        )

    def listLibraryObjects(self):
        """
        Return a list of library objects.

        Known limitation: library objects lacking an soname are not included.

        @rtype: list of strings
        @return: list of paths to all providers

        """
        rValue = []
        if not self._libs:
            self.rebuild()
        for arch_map in self._libs.values():
            for soname_map in arch_map.values():
                for obj_key in soname_map.providers:
                    rValue.extend(self._obj_properties[obj_key].alt_paths)
        return rValue

    def getOwners(self, obj):
        """
        Return the package(s) associated with an object. Raises KeyError
        if the object is unknown. Returns an empty tuple if the owner(s)
        are unknown.

        NOTE: For preserved libraries, the owner(s) may have been
        previously uninstalled, but these uninstalled owners can be
        returned by this method since they are registered in the
        PreservedLibsRegistry.

        @param obj: absolute path to an object
        @type obj: string (example: '/usr/bin/bar')
        @rtype: tuple
        @return: a tuple of cpv
        """
        if not self._libs:
            self.rebuild()
        if isinstance(obj, self._ObjectKey):
            obj_key = obj
        else:
            obj_key = self._obj_key_cache.get(obj)
            if obj_key is None:
                raise KeyError("%s not in object list" % obj)
        obj_props = self._obj_properties.get(obj_key)
        if obj_props is None:
            raise KeyError("%s not in object list" % obj_key)
        if obj_props.owner is None:
            return ()
        return (obj_props.owner,)

    def getSoname(self, obj):
        """
        Return the soname associated with an object.

        @param obj: absolute path to an object
        @type obj: string (example: '/usr/bin/bar')
        @rtype: string
        @return: soname as a string

        """
        if not self._libs:
            self.rebuild()
        if isinstance(obj, self._ObjectKey):
            obj_key = obj
            if obj_key not in self._obj_properties:
                raise KeyError("%s not in object list" % obj_key)
            return self._obj_properties[obj_key].soname
        if obj not in self._obj_key_cache:
            raise KeyError("%s not in object list" % obj)
        return self._obj_properties[self._obj_key_cache[obj]].soname

    def findProviders(self, obj):
        """
        Find providers for an object or object key.

        This method may be called with a key from _obj_properties.

        In some cases, not all valid libraries are returned.  This may occur when
        an soname symlink referencing a library is in an object's runpath while
        the actual library is not.  We should consider cataloging symlinks within
        LinkageMap as this would avoid those cases and would be a better model of
        library dependencies (since the dynamic linker actually searches for
        files named with the soname in the runpaths).

        @param obj: absolute path to an object or a key from _obj_properties
        @type obj: string (example: '/usr/bin/bar') or _ObjectKey
        @rtype: dict (example: {'libbar.so': set(['/lib/libbar.so.1.5'])})
        @return: The return value is a soname -> set-of-library-paths, where
        set-of-library-paths satisfy soname.

        """

        os = _os_merge

        rValue = {}

        if not self._libs:
            self.rebuild()

        # Determine the obj_key from the arguments.
        if isinstance(obj, self._ObjectKey):
            obj_key = obj
            if obj_key not in self._obj_properties:
                raise KeyError("%s not in object list" % obj_key)
        else:
            obj_key = self._obj_key(obj)
            if obj_key not in self._obj_properties:
                raise KeyError("{} ({}) not in object list".format(obj_key, obj))

        obj_props = self._obj_properties[obj_key]
        arch = obj_props.arch
        needed = obj_props.needed
        path = obj_props.runpaths
        path_keys = {self._path_key(x) for x in path.union(self._defpath)}
        for soname in needed:
            rValue[soname] = set()
            if arch not in self._libs or soname not in self._libs[arch]:
                continue
            # For each potential provider of the soname, add it to rValue if it
            # resides in the obj's runpath.
            for provider_key in self._libs[arch][soname].providers:
                providers = self._obj_properties[provider_key].alt_paths
                for provider in providers:
                    if self._path_key(os.path.dirname(provider)) in path_keys:
                        rValue[soname].add(provider)
        return rValue

    def findConsumers(self, obj, exclude_providers=None, greedy=True):
        """
        Find consumers of an object or object key.

        This method may be called with a key from _obj_properties.  If this
        method is going to be called with an object key, to avoid not catching
        shadowed libraries, do not pass new _ObjectKey instances to this method.
        Instead pass the obj as a string.

        In some cases, not all consumers are returned.  This may occur when
        an soname symlink referencing a library is in an object's runpath while
        the actual library is not. For example, this problem is noticeable for
        binutils since it's libraries are added to the path via symlinks that
        are gemerated in the /usr/$CHOST/lib/ directory by binutils-config.
        Failure to recognize consumers of these symlinks makes preserve-libs
        fail to preserve binutils libs that are needed by these unrecognized
        consumers.

        Note that library consumption via dlopen (common for kde plugins) is
        currently undetected. However, it is possible to use the
        corresponding libtool archive (*.la) files to detect such consumers
        (revdep-rebuild is able to detect them).

        The exclude_providers argument is useful for determining whether
        removal of one or more packages will create unsatisfied consumers. When
        this option is given, consumers are excluded from the results if there
        is an alternative provider (which is not excluded) of the required
        soname such that the consumers will remain satisfied if the files
        owned by exclude_providers are removed.

        @param obj: absolute path to an object or a key from _obj_properties
        @type obj: string (example: '/usr/bin/bar') or _ObjectKey
        @param exclude_providers: A collection of callables that each take a
                single argument referring to the path of a library (example:
                '/usr/lib/libssl.so.0.9.8'), and return True if the library is
                owned by a provider which is planned for removal.
        @type exclude_providers: collection
        @param greedy: If True, then include consumers that are satisfied
        by alternative providers, otherwise omit them. Default is True.
        @type greedy: Boolean
        @rtype: set of strings (example: set(['/bin/foo', '/usr/bin/bar']))
        @return: The return value is a soname -> set-of-library-paths, where
        set-of-library-paths satisfy soname.

        """

        os = _os_merge

        if not self._libs:
            self.rebuild()

        # Determine the obj_key and the set of objects matching the arguments.
        if isinstance(obj, self._ObjectKey):
            obj_key = obj
            if obj_key not in self._obj_properties:
                raise KeyError("%s not in object list" % obj_key)
            objs = self._obj_properties[obj_key].alt_paths
        else:
            objs = {obj}
            obj_key = self._obj_key(obj)
            if obj_key not in self._obj_properties:
                raise KeyError("{} ({}) not in object list".format(obj_key, obj))

        # If there is another version of this lib with the
        # same soname and the soname symlink points to that
        # other version, this lib will be shadowed and won't
        # have any consumers.
        if not isinstance(obj, self._ObjectKey):
            soname = self._obj_properties[obj_key].soname
            soname_link = os.path.join(
                self._root, os.path.dirname(obj).lstrip(os.path.sep), soname
            )
            obj_path = os.path.join(self._root, obj.lstrip(os.sep))
            try:
                soname_st = os.stat(soname_link)
                obj_st = os.stat(obj_path)
            except OSError:
                pass
            else:
                if (obj_st.st_dev, obj_st.st_ino) != (
                    soname_st.st_dev,
                    soname_st.st_ino,
                ):
                    return set()

        obj_props = self._obj_properties[obj_key]
        arch = obj_props.arch
        soname = obj_props.soname

        soname_node = None
        arch_map = self._libs.get(arch)
        if arch_map is not None:
            soname_node = arch_map.get(soname)

        defpath_keys = {self._path_key(x) for x in self._defpath}
        satisfied_consumer_keys = set()
        if soname_node is not None:
            if exclude_providers is not None or not greedy:
                relevant_dir_keys = set()
                for provider_key in soname_node.providers:
                    if not greedy and provider_key == obj_key:
                        continue
                    provider_objs = self._obj_properties[provider_key].alt_paths
                    for p in provider_objs:
                        provider_excluded = False
                        if exclude_providers is not None:
                            for excluded_provider_isowner in exclude_providers:
                                if excluded_provider_isowner(p):
                                    provider_excluded = True
                                    break
                        if not provider_excluded:
                            # This provider is not excluded. It will
                            # satisfy a consumer of this soname if it
                            # is in the default ld.so path or the
                            # consumer's runpath.
                            relevant_dir_keys.add(self._path_key(os.path.dirname(p)))

                if relevant_dir_keys:
                    for consumer_key in soname_node.consumers:
                        path = self._obj_properties[consumer_key].runpaths
                        path_keys = defpath_keys.copy()
                        path_keys.update(self._path_key(x) for x in path)
                        if relevant_dir_keys.intersection(path_keys):
                            satisfied_consumer_keys.add(consumer_key)

        rValue = set()
        if soname_node is not None:
            # For each potential consumer, add it to rValue if an object from the
            # arguments resides in the consumer's runpath.
            objs_dir_keys = {self._path_key(os.path.dirname(x)) for x in objs}
            for consumer_key in soname_node.consumers:
                if consumer_key in satisfied_consumer_keys:
                    continue
                consumer_props = self._obj_properties[consumer_key]
                path = consumer_props.runpaths
                consumer_objs = consumer_props.alt_paths
                path_keys = defpath_keys.union(self._path_key(x) for x in path)
                if objs_dir_keys.intersection(path_keys):
                    rValue.update(consumer_objs)
        return rValue



# Copyright 2012-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys
import textwrap

import portage
from portage import os
from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs
from portage._global_updates import _do_global_updates
from portage.output import colorize


class MoveSlotEntTestCase(TestCase):
    def testMoveSlotEnt(self):

        ebuilds = {
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "5",
                "SLOT": "0/2.30",
            },
            "dev-libs/B-2::dont_apply_updates": {
                "SLOT": "0",
            },
            "dev-libs/C-2.1::dont_apply_updates": {
                "EAPI": "5",
                "SLOT": "0/2.1",
            },
        }

        installed = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "5",
                "SLOT": "0/2.30",
            },
            "dev-libs/B-1::test_repo": {
                "SLOT": "0",
            },
            "dev-libs/C-1::test_repo": {
                "EAPI": "5",
                "SLOT": "0/1",
            },
        }

        binpkgs = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "5",
                "SLOT": "0/2.30",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "5",
                "SLOT": "0/2.30",
            },
            "dev-libs/B-1::test_repo": {
                "SLOT": "0",
            },
            "dev-libs/B-2::dont_apply_updates": {
                "SLOT": "0",
            },
            "dev-libs/C-1::test_repo": {
                "EAPI": "5",
                "SLOT": "0/1",
            },
            "dev-libs/C-2.1::dont_apply_updates": {
                "EAPI": "5",
                "SLOT": "0/2.1",
            },
        }

        updates = textwrap.dedent(
            """
			slotmove dev-libs/A 0 2
			slotmove dev-libs/B 0 1
			slotmove dev-libs/C 0 1
		"""
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                settings = playground.settings
                trees = playground.trees
                eroot = settings["EROOT"]
                test_repo_location = settings.repositories["test_repo"].location
                portdb = trees[eroot]["porttree"].dbapi
                vardb = trees[eroot]["vartree"].dbapi
                bindb = trees[eroot]["bintree"].dbapi

                updates_dir = os.path.join(test_repo_location, "profiles", "updates")

                try:
                    ensure_dirs(updates_dir)
                    with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
                        f.write(updates)

                    # Create an empty updates directory, so that this
                    # repo doesn't inherit updates from the main repo.
                    ensure_dirs(
                        os.path.join(
                            portdb.getRepositoryPath("dont_apply_updates"),
                            "profiles",
                            "updates",
                        )
                    )

                    global_noiselimit = portage.util.noiselimit
                    portage.util.noiselimit = -2
                    try:
                        _do_global_updates(trees, {})
                    finally:
                        portage.util.noiselimit = global_noiselimit

                    # Workaround for cache validation not working
                    # correctly when filesystem has timestamp precision
                    # of 1 second.
                    vardb._clear_cache()

                    # 0/2.30 -> 2/2.30
                    self.assertEqual(
                        "2/2.30", vardb.aux_get("dev-libs/A-1", ["SLOT"])[0]
                    )
                    self.assertEqual(
                        "2/2.30", bindb.aux_get("dev-libs/A-1", ["SLOT"])[0]
                    )

                    # 0 -> 1
                    self.assertEqual("1", vardb.aux_get("dev-libs/B-1", ["SLOT"])[0])
                    self.assertEqual("1", bindb.aux_get("dev-libs/B-1", ["SLOT"])[0])

                    # 0/1 -> 1 (equivalent to 1/1)
                    self.assertEqual("1", vardb.aux_get("dev-libs/C-1", ["SLOT"])[0])
                    self.assertEqual("1", bindb.aux_get("dev-libs/C-1", ["SLOT"])[0])

                    # dont_apply_updates
                    self.assertEqual(
                        "0/2.30", bindb.aux_get("dev-libs/A-2", ["SLOT"])[0]
                    )
                    self.assertEqual("0", bindb.aux_get("dev-libs/B-2", ["SLOT"])[0])
                    self.assertEqual(
                        "0/2.1", bindb.aux_get("dev-libs/C-2.1", ["SLOT"])[0]
                    )

                finally:
                    playground.cleanup()



# Copyright 2011-2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import tempfile

import portage
from portage import os
from portage import shutil
from portage.dbapi.virtual import fakedbapi
from portage.package.ebuild.config import config
from portage.tests import TestCase


class TestFakedbapi(TestCase):
    def testFakedbapi(self):
        packages = (
            (
                "app-misc/foo-1",
                {
                    "EAPI": "2",  # does not support IUSE_EFFECTIVE
                    "IUSE": "",
                    "repository": "gentoo",
                    "SLOT": "1",
                    "USE": "missing-iuse",
                },
            ),
            (
                "app-misc/foo-2",
                {
                    "EAPI": "5",  # supports IUSE_EFFECTIVE
                    "IUSE": "",
                    "repository": "gentoo",
                    "SLOT": "2",
                    "USE": "missing-iuse",
                },
            ),
            (
                "sys-apps/portage-2.1.10",
                {
                    "EAPI": "2",
                    "IUSE": "ipc doc",
                    "repository": "gentoo",
                    "SLOT": "0",
                    "USE": "ipc missing-iuse",
                },
            ),
            (
                "virtual/package-manager-0",
                {
                    "EAPI": "0",
                    "repository": "gentoo",
                    "SLOT": "0",
                },
            ),
        )

        match_tests = (
            # The missing-iuse match is only intended to work for binary
            # packages with EAPIs that support IUSE_EFFECTIVE (bug 640318).
            ("app-misc/foo[missing-iuse]", ["app-misc/foo-2"]),
            ("app-misc/foo[-missing-iuse]", []),
            ("app-misc/foo", ["app-misc/foo-1", "app-misc/foo-2"]),
            ("sys-apps/portage:0[ipc]", ["sys-apps/portage-2.1.10"]),
            ("sys-apps/portage:0[-ipc]", []),
            ("sys-apps/portage:0[doc]", []),
            ("sys-apps/portage:0[-doc]", ["sys-apps/portage-2.1.10"]),
            ("sys-apps/portage:0", ["sys-apps/portage-2.1.10"]),
            ("sys-apps/portage:0[missing-iuse]", []),
            ("sys-apps/portage:0[-missing-iuse]", []),
            ("sys-apps/portage:0::gentoo[ipc]", ["sys-apps/portage-2.1.10"]),
            ("sys-apps/portage:0::multilib[ipc]", []),
            ("virtual/package-manager", ["virtual/package-manager-0"]),
        )

        tempdir = tempfile.mkdtemp()
        try:
            test_repo = os.path.join(tempdir, "var", "repositories", "test_repo")
            os.makedirs(os.path.join(test_repo, "profiles"))
            with open(os.path.join(test_repo, "profiles", "repo_name"), "w") as f:
                f.write("test_repo")
            env = {
                "PORTAGE_REPOSITORIES": "[DEFAULT]\nmain-repo = test_repo\n[test_repo]\nlocation = %s"
                % test_repo
            }

            # Tests may override portage.const.EPREFIX in order to
            # simulate a prefix installation. It's reasonable to do
            # this because tests should be self-contained such that
            # the "real" value of portage.const.EPREFIX is entirely
            # irrelevant (see bug #492932).
            portage.const.EPREFIX = tempdir

            fakedb = fakedbapi(
                settings=config(config_profile_path="", env=env, eprefix=tempdir)
            )
            for cpv, metadata in packages:
                fakedb.cpv_inject(cpv, metadata=metadata)

            for atom, expected_result in match_tests:
                result = fakedb.match(atom)
                self.assertEqual(
                    fakedb.match(atom),
                    expected_result,
                    "fakedb.match('{}') = {} != {}".format(
                        atom, result, expected_result
                    ),
                )
        finally:
            shutil.rmtree(tempdir)



# Copyright 2011-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import subprocess
import sys

import portage
from portage import shutil, os
from portage.const import (
    BASH_BINARY,
    BINREPOS_CONF_FILE,
    PORTAGE_PYM_PATH,
    USER_CONFIG_PATH,
    SUPPORTED_GENTOO_BINPKG_FORMATS,
)
from portage.cache.mappings import Mapping
from portage.process import find_binary
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.tests.util.test_socks5 import AsyncHTTPServer
from portage.util import ensure_dirs, find_updated_config_files, shlex_split
from portage.util.futures import asyncio
from portage.output import colorize


class BinhostContentMap(Mapping):
    def __init__(self, remote_path, local_path):
        self._remote_path = remote_path
        self._local_path = local_path

    def __getitem__(self, request_path):
        safe_path = os.path.normpath(request_path)
        if not safe_path.startswith(self._remote_path + "/"):
            raise KeyError(request_path)
        local_path = os.path.join(
            self._local_path, safe_path[len(self._remote_path) + 1 :]
        )
        try:
            with open(local_path, "rb") as f:
                return f.read()
        except OSError:
            raise KeyError(request_path)


class SimpleEmergeTestCase(TestCase):
    def _have_python_xml(self):
        try:
            __import__("xml.etree.ElementTree")
            __import__("xml.parsers.expat").parsers.expat.ExpatError
        except (AttributeError, ImportError):
            return False
        return True

    def testSimple(self):

        debug = False

        install_something = """
S="${WORKDIR}"

pkg_pretend() {
	einfo "called pkg_pretend for $CATEGORY/$PF"
}

src_install() {
	einfo "installing something..."
	insinto /usr/lib/${P}
	echo "blah blah blah" > "${T}"/regular-file
	doins "${T}"/regular-file
	dosym regular-file /usr/lib/${P}/symlink || die

	# Test CONFIG_PROTECT
	insinto /etc
	newins "${T}"/regular-file ${PN}-${SLOT%/*}

	# Test code for bug #381629, using a copyright symbol encoded with latin-1.
	# We use $(printf "\\xa9") rather than $'\\xa9', since printf apparently
	# works in any case, while $'\\xa9' transforms to \\xef\\xbf\\xbd under
	# some conditions. TODO: Find out why it transforms to \\xef\\xbf\\xbd when
	# running tests for Python 3.2 (even though it's bash that is ultimately
	# responsible for performing the transformation).
	local latin_1_dir=/usr/lib/${P}/latin-1-$(printf "\\xa9")-directory
	insinto "${latin_1_dir}"
	echo "blah blah blah" > "${T}"/latin-1-$(printf "\\xa9")-regular-file || die
	doins "${T}"/latin-1-$(printf "\\xa9")-regular-file
	dosym latin-1-$(printf "\\xa9")-regular-file ${latin_1_dir}/latin-1-$(printf "\\xa9")-symlink || die

	call_has_and_best_version
}

pkg_config() {
	einfo "called pkg_config for $CATEGORY/$PF"
}

pkg_info() {
	einfo "called pkg_info for $CATEGORY/$PF"
}

pkg_preinst() {
	if ! ___eapi_best_version_and_has_version_support_-b_-d_-r; then
		# The BROOT variable is unset during pkg_* phases for EAPI 7,
		# therefore best/has_version -b is expected to fail if we attempt
		# to call it for EAPI 7 here.
		call_has_and_best_version
	fi
}

call_has_and_best_version() {
	local root_arg
	if ___eapi_best_version_and_has_version_support_-b_-d_-r; then
		root_arg="-b"
	else
		root_arg="--host-root"
	fi
	einfo "called ${EBUILD_PHASE_FUNC} for $CATEGORY/$PF"
	einfo "EPREFIX=${EPREFIX}"
	einfo "PORTAGE_OVERRIDE_EPREFIX=${PORTAGE_OVERRIDE_EPREFIX}"
	einfo "ROOT=${ROOT}"
	einfo "EROOT=${EROOT}"
	einfo "SYSROOT=${SYSROOT}"
	einfo "ESYSROOT=${ESYSROOT}"
	einfo "BROOT=${BROOT}"
	# Test that has_version and best_version work correctly with
	# prefix (involves internal ROOT -> EROOT calculation in order
	# to support ROOT override via the environment with EAPIs 3
	# and later which support prefix).
	if has_version $CATEGORY/$PN:$SLOT ; then
		einfo "has_version detects an installed instance of $CATEGORY/$PN:$SLOT"
		einfo "best_version reports that the installed instance is $(best_version $CATEGORY/$PN:$SLOT)"
	else
		einfo "has_version does not detect an installed instance of $CATEGORY/$PN:$SLOT"
	fi
	if [[ ${EPREFIX} != ${PORTAGE_OVERRIDE_EPREFIX} ]] ; then
		if has_version ${root_arg} $CATEGORY/$PN:$SLOT ; then
			einfo "has_version ${root_arg} detects an installed instance of $CATEGORY/$PN:$SLOT"
			einfo "best_version ${root_arg} reports that the installed instance is $(best_version ${root_arg} $CATEGORY/$PN:$SLOT)"
		else
			einfo "has_version ${root_arg} does not detect an installed instance of $CATEGORY/$PN:$SLOT"
		fi
	fi
}

"""

        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "IUSE": "+flag",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "MISC_CONTENT": install_something,
                "RDEPEND": "flag? ( dev-libs/B[flag] )",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "IUSE": "+flag",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "MISC_CONTENT": install_something,
            },
            "dev-libs/C-1": {
                "EAPI": "7",
                "KEYWORDS": "~x86",
                "RDEPEND": "dev-libs/D[flag]",
                "MISC_CONTENT": install_something,
            },
            "dev-libs/D-1": {
                "EAPI": "7",
                "KEYWORDS": "~x86",
                "IUSE": "flag",
                "MISC_CONTENT": install_something,
            },
            "virtual/foo-0": {
                "EAPI": "5",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
            },
        }

        installed = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "IUSE": "+flag",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "RDEPEND": "flag? ( dev-libs/B[flag] )",
                "USE": "flag",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "IUSE": "+flag",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "USE": "flag",
            },
            "dev-libs/depclean-me-1": {
                "EAPI": "5",
                "IUSE": "",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "USE": "",
            },
            "app-misc/depclean-me-1": {
                "EAPI": "5",
                "IUSE": "",
                "KEYWORDS": "x86",
                "LICENSE": "GPL-2",
                "RDEPEND": "dev-libs/depclean-me",
                "USE": "",
            },
        }

        metadata_xml_files = (
            (
                "dev-libs/A",
                {
                    "flags": "<flag name='flag'>Description of how USE='flag' affects this package</flag>",
                },
            ),
            (
                "dev-libs/B",
                {
                    "flags": "<flag name='flag'>Description of how USE='flag' affects this package</flag>",
                },
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    installed=installed,
                    debug=debug,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                loop = asyncio._wrap_loop()
                loop.run_until_complete(
                    asyncio.ensure_future(
                        self._async_test_simple(
                            playground, metadata_xml_files, loop=loop
                        ),
                        loop=loop,
                    )
                )

    async def _async_test_simple(self, playground, metadata_xml_files, loop):

        debug = playground.debug
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        trees = playground.trees
        portdb = trees[eroot]["porttree"].dbapi
        test_repo_location = settings.repositories["test_repo"].location
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        cachedir = os.path.join(var_cache_edb, "dep")
        cachedir_pregen = os.path.join(test_repo_location, "metadata", "md5-cache")

        portage_python = portage._python_interpreter
        dispatch_conf_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.sbindir, "dispatch-conf"),
        )
        ebuild_cmd = (portage_python, "-b", "-Wd", os.path.join(self.bindir, "ebuild"))
        egencache_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.bindir, "egencache"),
            "--repo",
            "test_repo",
            "--repositories-configuration",
            settings.repositories.config_string(),
        )
        emerge_cmd = (portage_python, "-b", "-Wd", os.path.join(self.bindir, "emerge"))
        emaint_cmd = (portage_python, "-b", "-Wd", os.path.join(self.sbindir, "emaint"))
        env_update_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.sbindir, "env-update"),
        )
        etc_update_cmd = (BASH_BINARY, os.path.join(self.sbindir, "etc-update"))
        fixpackages_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.sbindir, "fixpackages"),
        )
        portageq_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.bindir, "portageq"),
        )
        quickpkg_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.bindir, "quickpkg"),
        )
        regenworld_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.sbindir, "regenworld"),
        )

        rm_binary = find_binary("rm")
        self.assertEqual(rm_binary is None, False, "rm command not found")
        rm_cmd = (rm_binary,)

        egencache_extra_args = []
        if self._have_python_xml():
            egencache_extra_args.append("--update-use-local-desc")

        test_ebuild = portdb.findname("dev-libs/A-1")
        self.assertFalse(test_ebuild is None)

        cross_prefix = os.path.join(eprefix, "cross_prefix")
        cross_root = os.path.join(eprefix, "cross_root")
        cross_eroot = os.path.join(cross_root, eprefix.lstrip(os.sep))

        binhost_dir = os.path.join(eprefix, "binhost")
        binhost_address = "127.0.0.1"
        binhost_remote_path = "/binhost"
        binhost_server = AsyncHTTPServer(
            binhost_address, BinhostContentMap(binhost_remote_path, binhost_dir), loop
        ).__enter__()
        binhost_uri = "http://{address}:{port}{path}".format(
            address=binhost_address,
            port=binhost_server.server_port,
            path=binhost_remote_path,
        )

        binpkg_format = settings.get(
            "BINPKG_FORMAT", SUPPORTED_GENTOO_BINPKG_FORMATS[0]
        )
        self.assertIn(binpkg_format, ("xpak", "gpkg"))
        if binpkg_format == "xpak":
            foo_filename = "foo-0-1.xpak"
        elif binpkg_format == "gpkg":
            foo_filename = "foo-0-1.gpkg.tar"

        test_commands = ()

        if hasattr(argparse.ArgumentParser, "parse_intermixed_args"):
            test_commands += (
                emerge_cmd + ("--oneshot", "dev-libs/A", "-v", "dev-libs/A"),
            )

        test_commands += (
            emerge_cmd
            + (
                "--usepkgonly",
                "--root",
                cross_root,
                "--quickpkg-direct=y",
                "--quickpkg-direct-root",
                "/",
                "dev-libs/A",
            ),
            emerge_cmd
            + (
                "--usepkgonly",
                "--quickpkg-direct=y",
                "--quickpkg-direct-root",
                cross_root,
                "dev-libs/A",
            ),
            env_update_cmd,
            portageq_cmd
            + (
                "envvar",
                "-v",
                "CONFIG_PROTECT",
                "EROOT",
                "PORTAGE_CONFIGROOT",
                "PORTAGE_TMPDIR",
                "USERLAND",
            ),
            etc_update_cmd,
            dispatch_conf_cmd,
            emerge_cmd + ("--version",),
            emerge_cmd + ("--info",),
            emerge_cmd + ("--info", "--verbose"),
            emerge_cmd + ("--list-sets",),
            emerge_cmd + ("--check-news",),
            rm_cmd + ("-rf", cachedir),
            rm_cmd + ("-rf", cachedir_pregen),
            emerge_cmd + ("--regen",),
            rm_cmd + ("-rf", cachedir),
            ({"FEATURES": "metadata-transfer"},) + emerge_cmd + ("--regen",),
            rm_cmd + ("-rf", cachedir),
            ({"FEATURES": "metadata-transfer"},) + emerge_cmd + ("--regen",),
            rm_cmd + ("-rf", cachedir),
            egencache_cmd + ("--update",) + tuple(egencache_extra_args),
            ({"FEATURES": "metadata-transfer"},) + emerge_cmd + ("--metadata",),
            rm_cmd + ("-rf", cachedir),
            ({"FEATURES": "metadata-transfer"},) + emerge_cmd + ("--metadata",),
            emerge_cmd + ("--metadata",),
            rm_cmd + ("-rf", cachedir),
            emerge_cmd + ("--oneshot", "virtual/foo"),
            lambda: self.assertFalse(
                os.path.exists(os.path.join(pkgdir, "virtual", "foo", foo_filename))
            ),
            ({"FEATURES": "unmerge-backup"},)
            + emerge_cmd
            + ("--unmerge", "virtual/foo"),
            lambda: self.assertTrue(
                os.path.exists(os.path.join(pkgdir, "virtual", "foo", foo_filename))
            ),
            emerge_cmd + ("--pretend", "dev-libs/A"),
            ebuild_cmd + (test_ebuild, "manifest", "clean", "package", "merge"),
            emerge_cmd + ("--pretend", "--tree", "--complete-graph", "dev-libs/A"),
            emerge_cmd + ("-p", "dev-libs/B"),
            emerge_cmd + ("-p", "--newrepo", "dev-libs/B"),
            emerge_cmd
            + (
                "-B",
                "dev-libs/B",
            ),
            emerge_cmd
            + (
                "--oneshot",
                "--usepkg",
                "dev-libs/B",
            ),
            # trigger clean prior to pkg_pretend as in bug #390711
            ebuild_cmd + (test_ebuild, "unpack"),
            emerge_cmd
            + (
                "--oneshot",
                "dev-libs/A",
            ),
            emerge_cmd
            + (
                "--noreplace",
                "dev-libs/A",
            ),
            emerge_cmd
            + (
                "--config",
                "dev-libs/A",
            ),
            emerge_cmd + ("--info", "dev-libs/A", "dev-libs/B"),
            emerge_cmd + ("--pretend", "--depclean", "--verbose", "dev-libs/B"),
            emerge_cmd
            + (
                "--pretend",
                "--depclean",
            ),
            emerge_cmd + ("--depclean",),
            quickpkg_cmd
            + (
                "--include-config",
                "y",
                "dev-libs/A",
            ),
            # Test bug #523684, where a file renamed or removed by the
            # admin forces replacement files to be merged with config
            # protection.
            lambda: self.assertEqual(
                0,
                len(
                    list(
                        find_updated_config_files(
                            eroot, shlex_split(settings["CONFIG_PROTECT"])
                        )
                    )
                ),
            ),
            lambda: os.unlink(os.path.join(eprefix, "etc", "A-0")),
            emerge_cmd + ("--usepkgonly", "dev-libs/A"),
            lambda: self.assertEqual(
                1,
                len(
                    list(
                        find_updated_config_files(
                            eroot, shlex_split(settings["CONFIG_PROTECT"])
                        )
                    )
                ),
            ),
            emaint_cmd + ("--check", "all"),
            emaint_cmd + ("--fix", "all"),
            fixpackages_cmd,
            regenworld_cmd,
            portageq_cmd + ("match", eroot, "dev-libs/A"),
            portageq_cmd + ("best_visible", eroot, "dev-libs/A"),
            portageq_cmd + ("best_visible", eroot, "binary", "dev-libs/A"),
            portageq_cmd + ("contents", eroot, "dev-libs/A-1"),
            portageq_cmd
            + ("metadata", eroot, "ebuild", "dev-libs/A-1", "EAPI", "IUSE", "RDEPEND"),
            portageq_cmd
            + ("metadata", eroot, "binary", "dev-libs/A-1", "EAPI", "USE", "RDEPEND"),
            portageq_cmd
            + (
                "metadata",
                eroot,
                "installed",
                "dev-libs/A-1",
                "EAPI",
                "USE",
                "RDEPEND",
            ),
            portageq_cmd + ("owners", eroot, eroot + "usr"),
            emerge_cmd + ("-p", eroot + "usr"),
            emerge_cmd + ("-p", "--unmerge", "-q", eroot + "usr"),
            emerge_cmd + ("--unmerge", "--quiet", "dev-libs/A"),
            emerge_cmd + ("-C", "--quiet", "dev-libs/B"),
            # If EMERGE_DEFAULT_OPTS contains --autounmask=n, then --autounmask
            # must be specified with --autounmask-continue.
            ({"EMERGE_DEFAULT_OPTS": "--autounmask=n"},)
            + emerge_cmd
            + (
                "--autounmask",
                "--autounmask-continue",
                "dev-libs/C",
            ),
            # Verify that the above --autounmask-continue command caused
            # USE=flag to be applied correctly to dev-libs/D.
            portageq_cmd + ("match", eroot, "dev-libs/D[flag]"),
            # Test cross-prefix usage, including chpathtool for binpkgs.
            # EAPI 7
            ({"EPREFIX": cross_prefix},) + emerge_cmd + ("dev-libs/C",),
            ({"EPREFIX": cross_prefix},)
            + portageq_cmd
            + ("has_version", cross_prefix, "dev-libs/C"),
            ({"EPREFIX": cross_prefix},)
            + portageq_cmd
            + ("has_version", cross_prefix, "dev-libs/D"),
            ({"ROOT": cross_root},) + emerge_cmd + ("dev-libs/D",),
            portageq_cmd + ("has_version", cross_eroot, "dev-libs/D"),
            # EAPI 5
            ({"EPREFIX": cross_prefix},) + emerge_cmd + ("--usepkgonly", "dev-libs/A"),
            ({"EPREFIX": cross_prefix},)
            + portageq_cmd
            + ("has_version", cross_prefix, "dev-libs/A"),
            ({"EPREFIX": cross_prefix},)
            + portageq_cmd
            + ("has_version", cross_prefix, "dev-libs/B"),
            ({"EPREFIX": cross_prefix},) + emerge_cmd + ("-C", "--quiet", "dev-libs/B"),
            ({"EPREFIX": cross_prefix},) + emerge_cmd + ("-C", "--quiet", "dev-libs/A"),
            ({"EPREFIX": cross_prefix},) + emerge_cmd + ("dev-libs/A",),
            ({"EPREFIX": cross_prefix},)
            + portageq_cmd
            + ("has_version", cross_prefix, "dev-libs/A"),
            ({"EPREFIX": cross_prefix},)
            + portageq_cmd
            + ("has_version", cross_prefix, "dev-libs/B"),
            # Test ROOT support
            ({"ROOT": cross_root},) + emerge_cmd + ("dev-libs/B",),
            portageq_cmd + ("has_version", cross_eroot, "dev-libs/B"),
        )

        # Test binhost support if FETCHCOMMAND is available.
        binrepos_conf_file = os.path.join(os.sep, eprefix, BINREPOS_CONF_FILE)
        with open(binrepos_conf_file, "w") as f:
            f.write("[test-binhost]\n")
            f.write("sync-uri = {}\n".format(binhost_uri))
        fetchcommand = portage.util.shlex_split(playground.settings["FETCHCOMMAND"])
        fetch_bin = portage.process.find_binary(fetchcommand[0])
        if fetch_bin is not None:
            test_commands = test_commands + (
                lambda: os.rename(pkgdir, binhost_dir),
                emerge_cmd + ("-e", "--getbinpkgonly", "dev-libs/A"),
                lambda: shutil.rmtree(pkgdir),
                lambda: os.rename(binhost_dir, pkgdir),
                # Remove binrepos.conf and test PORTAGE_BINHOST.
                lambda: os.unlink(binrepos_conf_file),
                lambda: os.rename(pkgdir, binhost_dir),
                ({"PORTAGE_BINHOST": binhost_uri},)
                + emerge_cmd
                + ("-fe", "--getbinpkgonly", "dev-libs/A"),
                lambda: shutil.rmtree(pkgdir),
                lambda: os.rename(binhost_dir, pkgdir),
            )

        distdir = playground.distdir
        pkgdir = playground.pkgdir
        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")
        profile_path = settings.profile_path
        user_config_dir = os.path.join(os.sep, eprefix, USER_CONFIG_PATH)

        path = os.environ.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "CLEAN_DELAY": "0",
            "DISTDIR": distdir,
            "EMERGE_WARNING_DELAY": "0",
            "INFODIR": "",
            "INFOPATH": "",
            "PATH": path,
            "PKGDIR": pkgdir,
            "PORTAGE_INST_GID": str(portage.data.portage_gid),
            "PORTAGE_INST_UID": str(portage.data.portage_uid),
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PORTAGE_TMPDIR": portage_tmpdir,
            "PORTAGE_LOGDIR": portage_tmpdir,
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "__PORTAGE_TEST_PATH_OVERRIDE": fake_bin,
        }

        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                "__PORTAGE_TEST_HARDLINK_LOCKS"
            ]

        updates_dir = os.path.join(test_repo_location, "profiles", "updates")
        dirs = [
            cachedir,
            cachedir_pregen,
            cross_eroot,
            cross_prefix,
            distdir,
            fake_bin,
            portage_tmpdir,
            updates_dir,
            user_config_dir,
            var_cache_edb,
        ]
        etc_symlinks = ("dispatch-conf.conf", "etc-update.conf")
        # Override things that may be unavailable, or may have portability
        # issues when running tests in exotic environments.
        #   prepstrip - bug #447810 (bash read builtin EINTR problem)
        true_symlinks = ["find", "prepstrip", "sed", "scanelf"]
        true_binary = find_binary("true")
        self.assertEqual(true_binary is None, False, "true command not found")
        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(true_binary, os.path.join(fake_bin, x))
            for x in etc_symlinks:
                os.symlink(
                    os.path.join(self.cnf_etc_path, x), os.path.join(eprefix, "etc", x)
                )
            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")
            # non-empty system set keeps --depclean quiet
            with open(os.path.join(profile_path, "packages"), "w") as f:
                f.write("*dev-libs/token-system-pkg")
            for cp, xml_data in metadata_xml_files:
                with open(
                    os.path.join(test_repo_location, cp, "metadata.xml"), "w"
                ) as f:
                    f.write(playground.metadata_xml_template % xml_data)
            with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
                f.write(
                    """
slotmove =app-doc/pms-3 2 3
move dev-util/git dev-vcs/git
"""
                )

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for args in test_commands:

                if hasattr(args, "__call__"):
                    args()
                    continue

                if isinstance(args[0], dict):
                    local_env = env.copy()
                    local_env.update(args[0])
                    args = args[1:]
                else:
                    local_env = env

                proc = await asyncio.create_subprocess_exec(
                    *args, env=local_env, stderr=None, stdout=stdout
                )

                if debug:
                    await proc.wait()
                else:
                    output, _err = await proc.communicate()
                    await proc.wait()
                    if proc.returncode != os.EX_OK:
                        portage.writemsg(output)

                self.assertEqual(
                    os.EX_OK, proc.returncode, "emerge failed with args {}".format(args)
                )
        finally:
            binhost_server.__exit__(None, None, None)
            playground.cleanup()



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from collections import deque

from portage import os
from _emerge.CompositeTask import CompositeTask
from _emerge.AsynchronousTask import AsynchronousTask


class TaskSequence(CompositeTask):
    """
    A collection of tasks that executes sequentially. Each task
    must have a addExitListener() method that can be used as
    a means to trigger movement from one task to the next.
    """

    __slots__ = ("_task_queue",)

    def __init__(self, **kwargs):
        AsynchronousTask.__init__(self, **kwargs)
        self._task_queue = deque()

    def add(self, task):
        self._task_queue.append(task)

    def _start(self):
        self._start_next_task()

    def _cancel(self):
        self._task_queue.clear()
        CompositeTask._cancel(self)

    def _start_next_task(self):
        try:
            task = self._task_queue.popleft()
        except IndexError:
            self._current_task = None
            self.returncode = os.EX_OK
            self.wait()
            return

        self._start_task(task, self._task_exit_handler)

    def _task_exit_handler(self, task):
        if self._default_exit(task) != os.EX_OK:
            self.wait()
        elif self._task_queue:
            self._start_next_task()
        else:
            self._final_exit(task)
            self.wait()

    def __bool__(self):
        return bool(self._task_queue)

    def __len__(self):
        return len(self._task_queue)



# Copyright 1998-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import errno
import tempfile
import portage
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage.const import BASH_BINARY
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from _emerge.SpawnProcess import SpawnProcess


class SpawnTestCase(TestCase):
    def testLogfile(self):
        logfile = None
        try:
            fd, logfile = tempfile.mkstemp()
            os.close(fd)
            null_fd = os.open("/dev/null", os.O_RDWR)
            test_string = 2 * "blah blah blah\n"
            proc = SpawnProcess(
                args=[BASH_BINARY, "-c", "echo -n '%s'" % test_string],
                env={},
                fd_pipes={0: portage._get_stdin().fileno(), 1: null_fd, 2: null_fd},
                scheduler=global_event_loop(),
                logfile=logfile,
            )
            proc.start()
            os.close(null_fd)
            self.assertEqual(proc.wait(), os.EX_OK)
            f = open(
                _unicode_encode(logfile, encoding=_encodings["fs"], errors="strict"),
                encoding=_encodings["content"],
                errors="strict",
            )
            log_content = f.read()
            f.close()
            # When logging passes through a pty, this comparison will fail
            # unless the oflag terminal attributes have the termios.OPOST
            # bit disabled. Otherwise, tranformations such as \n -> \r\n
            # may occur.
            self.assertEqual(test_string, log_content)
        finally:
            if logfile:
                try:
                    os.unlink(logfile)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                    del e



# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.dbapi import dbapi
from portage.dbapi.dep_expand import dep_expand
from portage.versions import cpv_getkey, _pkg_str


class fakedbapi(dbapi):
    """A fake dbapi that allows consumers to inject/remove packages to/from it
    portage.settings is required to maintain the dbAPI.
    """

    def __init__(self, settings=None, exclusive_slots=True, multi_instance=False):
        """
        @param exclusive_slots: When True, injecting a package with SLOT
                metadata causes an existing package in the same slot to be
                automatically removed (default is True).
        @type exclusive_slots: Boolean
        @param multi_instance: When True, multiple instances with the
                same cpv may be stored simultaneously, as long as they are
                distinguishable (default is False).
        @type multi_instance: Boolean
        """
        self._exclusive_slots = exclusive_slots
        self.cpvdict = {}
        self.cpdict = {}
        if settings is None:
            from portage import settings
        self.settings = settings
        self._match_cache = {}
        self._set_multi_instance(multi_instance)

    def _set_multi_instance(self, multi_instance):
        """
        Enable or disable multi_instance mode. This should before any
        packages are injected, so that all packages are indexed with
        the same implementation of self._instance_key.
        """
        if self.cpvdict:
            raise AssertionError(
                "_set_multi_instance called after " "packages have already been added"
            )
        self._multi_instance = multi_instance
        if multi_instance:
            self._instance_key = self._instance_key_multi_instance
        else:
            self._instance_key = self._instance_key_cpv

    def _instance_key_cpv(self, cpv, support_string=False):
        return cpv

    def _instance_key_multi_instance(self, cpv, support_string=False):
        try:
            return (cpv, cpv.build_id, cpv.file_size, cpv.build_time, cpv.mtime)
        except AttributeError:
            if not support_string:
                raise

        # Fallback for interfaces such as aux_get where API consumers
        # may pass in a plain string.
        latest = None
        for pkg in self.cp_list(cpv_getkey(cpv)):
            if pkg == cpv and (latest is None or latest.build_time < pkg.build_time):
                latest = pkg

        if latest is not None:
            return (
                latest,
                latest.build_id,
                latest.file_size,
                latest.build_time,
                latest.mtime,
            )

        raise KeyError(cpv)

    def clear(self):
        """
        Remove all packages.
        """
        self._clear_cache()
        self.cpvdict.clear()
        self.cpdict.clear()

    def _clear_cache(self):
        if self._categories is not None:
            self._categories = None
        if self._match_cache:
            self._match_cache = {}

    def match(self, origdep, use_cache=1):
        atom = dep_expand(origdep, mydb=self, settings=self.settings)
        cache_key = (atom, atom.unevaluated_atom)
        result = self._match_cache.get(cache_key)
        if result is not None:
            return result[:]
        result = list(self._iter_match(atom, self.cp_list(atom.cp)))
        self._match_cache[cache_key] = result
        return result[:]

    def cpv_exists(self, mycpv, myrepo=None):
        try:
            return self._instance_key(mycpv, support_string=True) in self.cpvdict
        except KeyError:
            # _instance_key failure
            return False

    def cp_list(self, mycp, use_cache=1, myrepo=None):
        # NOTE: Cache can be safely shared with the match cache, since the
        # match cache uses the result from dep_expand for the cache_key.
        cache_key = (mycp, mycp)
        cachelist = self._match_cache.get(cache_key)
        if cachelist is not None:
            return cachelist[:]
        cpv_list = self.cpdict.get(mycp)
        if cpv_list is None:
            cpv_list = []
        self._cpv_sort_ascending(cpv_list)
        self._match_cache[cache_key] = cpv_list
        return cpv_list[:]

    def cp_all(self, sort=False):
        return sorted(self.cpdict) if sort else list(self.cpdict)

    def cpv_all(self):
        if self._multi_instance:
            return [x[0] for x in self.cpvdict]
        return list(self.cpvdict)

    def cpv_inject(self, mycpv, metadata=None):
        """Adds a cpv to the list of available packages. See the
        exclusive_slots constructor parameter for behavior with
        respect to SLOT metadata.
        @param mycpv: cpv for the package to inject
        @type mycpv: str
        @param metadata: dictionary of raw metadata for aux_get() calls
        @param metadata: dict
        """
        self._clear_cache()

        try:
            mycp = mycpv.cp
        except AttributeError:
            mycp = None
        try:
            myslot = mycpv.slot
        except AttributeError:
            myslot = None

        if mycp is None or (
            myslot is None and metadata is not None and metadata.get("SLOT")
        ):
            if metadata is None:
                mycpv = _pkg_str(mycpv, db=self)
            else:
                mycpv = _pkg_str(
                    mycpv, metadata=metadata, settings=self.settings, db=self
                )

            mycp = mycpv.cp
            try:
                myslot = mycpv.slot
            except AttributeError:
                pass

        instance_key = self._instance_key(mycpv)
        self.cpvdict[instance_key] = metadata
        if not self._exclusive_slots:
            myslot = None
        if myslot and mycp in self.cpdict:
            # If necessary, remove another package in the same SLOT.
            for cpv in self.cpdict[mycp]:
                if instance_key != self._instance_key(cpv):
                    try:
                        other_slot = cpv.slot
                    except AttributeError:
                        pass
                    else:
                        if myslot == other_slot:
                            self.cpv_remove(cpv)
                            break

        cp_list = self.cpdict.get(mycp, [])
        cp_list = [x for x in cp_list if self._instance_key(x) != instance_key]
        cp_list.append(mycpv)
        self.cpdict[mycp] = cp_list

    def cpv_remove(self, mycpv):
        """Removes a cpv from the list of available packages."""
        self._clear_cache()
        mycp = cpv_getkey(mycpv)
        instance_key = self._instance_key(mycpv)
        self.cpvdict.pop(instance_key, None)
        cp_list = self.cpdict.get(mycp)
        if cp_list is not None:
            cp_list = [x for x in cp_list if self._instance_key(x) != instance_key]
            if cp_list:
                self.cpdict[mycp] = cp_list
            else:
                del self.cpdict[mycp]

    def aux_get(self, mycpv, wants, myrepo=None):
        metadata = self.cpvdict.get(self._instance_key(mycpv, support_string=True))
        if metadata is None:
            raise KeyError(mycpv)
        return [metadata.get(x, "") for x in wants]

    def aux_update(self, cpv, values):
        self._clear_cache()
        metadata = self.cpvdict.get(self._instance_key(cpv, support_string=True))
        if metadata is None:
            raise KeyError(cpv)
        metadata.update(values)


class testdbapi:
    """A dbapi instance with completely fake functions to get by hitting disk
    TODO(antarus):
    This class really needs to be rewritten to have better stubs; but these work for now.
    The dbapi classes themselves need unit tests...and that will be a lot of work.
    """

    def __init__(self):
        self.cpvs = {}

        def f(*args, **kwargs):
            return True

        fake_api = dir(dbapi)
        for call in fake_api:
            if not hasattr(self, call):
                setattr(self, call, f)



# Copyright 2012-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictRebuildTestCase(TestCase):
    def testSlotConflictRebuild(self):

        ebuilds = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/B-0": {
                "EAPI": "5",
                "DEPEND": "app-misc/A:=",
                "RDEPEND": "app-misc/A:=",
            },
            "app-misc/C-0": {
                "EAPI": "5",
                "DEPEND": "<app-misc/A-2",
                "RDEPEND": "<app-misc/A-2",
            },
            "app-misc/D-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/D-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/E-0": {
                "EAPI": "5",
                "DEPEND": "app-misc/D:=",
                "RDEPEND": "app-misc/D:=",
            },
        }

        installed = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/B-0": {
                "EAPI": "5",
                "DEPEND": "app-misc/A:0/1=",
                "RDEPEND": "app-misc/A:0/1=",
            },
            "app-misc/C-0": {
                "EAPI": "5",
                "DEPEND": "<app-misc/A-2",
                "RDEPEND": "<app-misc/A-2",
            },
            "app-misc/D-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/E-0": {
                "EAPI": "5",
                "DEPEND": "app-misc/D:0/1=",
                "RDEPEND": "app-misc/D:0/1=",
            },
        }

        world = ["app-misc/B", "app-misc/C", "app-misc/E"]

        test_cases = (
            # Test bug #439688, where a slot conflict prevents an
            # upgrade and we don't want to trigger unnecessary rebuilds.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--backtrack": 4},
                success=True,
                mergelist=["app-misc/D-2", "app-misc/E-0"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictMassRebuild(self):
        """
        Bug 486580
        Before this bug was fixed, emerge would backtrack for each package that needs
        a rebuild. This could cause it to hit the backtrack limit and not rebuild all
        needed packages.
        """
        ebuilds = {
            "app-misc/A-1": {
                "EAPI": "5",
                "DEPEND": "app-misc/B:=",
                "RDEPEND": "app-misc/B:=",
            },
            "app-misc/B-1": {"EAPI": "5", "SLOT": "1"},
            "app-misc/B-2": {"EAPI": "5", "SLOT": "2/2"},
        }

        installed = {
            "app-misc/B-1": {"EAPI": "5", "SLOT": "1"},
        }

        expected_mergelist = ["app-misc/A-1", "app-misc/B-2"]

        for i in range(5):
            ebuilds["app-misc/C%sC-1" % i] = {
                "EAPI": "5",
                "DEPEND": "app-misc/B:=",
                "RDEPEND": "app-misc/B:=",
            }

            installed["app-misc/C%sC-1" % i] = {
                "EAPI": "5",
                "DEPEND": "app-misc/B:1/1=",
                "RDEPEND": "app-misc/B:1/1=",
            }
            for x in ("DEPEND", "RDEPEND"):
                ebuilds["app-misc/A-1"][x] += " app-misc/C%sC" % i

            expected_mergelist.append("app-misc/C%sC-1" % i)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                ignore_mergelist_order=True,
                all_permutations=True,
                options={"--backtrack": 3, "--update": True, "--deep": True},
                success=True,
                mergelist=expected_mergelist,
            ),
        )

        world = []

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictForgottenChild(self):
        """
        Similar to testSlotConflictMassRebuild above, but this time the rebuilds are scheduled,
        but the package causing the rebuild (the child) is not installed.
        """
        ebuilds = {
            "app-misc/A-2": {
                "EAPI": "5",
                "DEPEND": "app-misc/B:= app-misc/C",
                "RDEPEND": "app-misc/B:= app-misc/C",
            },
            "app-misc/B-2": {"EAPI": "5", "SLOT": "2"},
            "app-misc/C-1": {
                "EAPI": "5",
                "DEPEND": "app-misc/B:=",
                "RDEPEND": "app-misc/B:=",
            },
        }

        installed = {
            "app-misc/A-1": {
                "EAPI": "5",
                "DEPEND": "app-misc/B:1/1= app-misc/C",
                "RDEPEND": "app-misc/B:1/1= app-misc/C",
            },
            "app-misc/B-1": {"EAPI": "5", "SLOT": "1"},
            "app-misc/C-1": {
                "EAPI": "5",
                "DEPEND": "app-misc/B:1/1=",
                "RDEPEND": "app-misc/B:1/1=",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"], success=True, mergelist=["app-misc/A-2"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["app-misc/B-2", "app-misc/C-1", "app-misc/A-2"],
            ),
        )

        world = []

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictDepChange(self):
        """
        Bug 490362
        The dependency in the ebuild was changed form slot operator to
        no slot operator. The vdb contained the slot operator and emerge
        would refuse to rebuild.
        """
        ebuilds = {
            "app-misc/A-1": {
                "EAPI": "5",
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
            },
            "app-misc/B-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/B-2": {"EAPI": "5", "SLOT": "0/2"},
        }

        installed = {
            "app-misc/A-1": {
                "EAPI": "5",
                "DEPEND": "app-misc/B:0/1=",
                "RDEPEND": "app-misc/B:0/1=",
            },
            "app-misc/B-1": {"EAPI": "5", "SLOT": "0/1"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/B"], success=True, mergelist=["app-misc/B-2", "app-misc/A-1"]
            ),
        )

        world = ["app-misc/A"]

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictMixedDependencies(self):
        """
        Bug 487198
        For parents with mixed >= and < dependencies, we scheduled rebuilds for the
        >= atom, but in the end didn't install the child update because of the < atom.
        """
        ebuilds = {
            "cat/slotted-lib-1": {"EAPI": "5", "SLOT": "1"},
            "cat/slotted-lib-2": {"EAPI": "5", "SLOT": "2"},
            "cat/slotted-lib-3": {"EAPI": "5", "SLOT": "3"},
            "cat/slotted-lib-4": {"EAPI": "5", "SLOT": "4"},
            "cat/slotted-lib-5": {"EAPI": "5", "SLOT": "5"},
            "cat/user-1": {
                "EAPI": "5",
                "DEPEND": ">=cat/slotted-lib-2:= <cat/slotted-lib-4:=",
                "RDEPEND": ">=cat/slotted-lib-2:= <cat/slotted-lib-4:=",
            },
        }

        installed = {
            "cat/slotted-lib-3": {"EAPI": "5", "SLOT": "3"},
            "cat/user-1": {
                "EAPI": "5",
                "DEPEND": ">=cat/slotted-lib-2:3/3= <cat/slotted-lib-4:3/3=",
                "RDEPEND": ">=cat/slotted-lib-2:3/3= <cat/slotted-lib-4:3/3=",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["cat/user"],
                options={"--deep": True, "--update": True},
                success=True,
                mergelist=[],
            ),
        )

        world = []

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictMultiRepo(self):
        """
        Bug 497238
        Different repositories contain the same cpv with different sub-slots for
        a slot operator child.
        Downgrading the slot operator parent would result in a sub-slot change of
        the installed package by changing the source repository.
        Make sure we don't perform this undesirable rebuild.
        """
        ebuilds = {
            "net-firewall/iptables-1.4.21::overlay": {"EAPI": "5", "SLOT": "0/10"},
            "sys-apps/iproute2-3.11.0::overlay": {
                "EAPI": "5",
                "RDEPEND": "net-firewall/iptables:=",
            },
            "net-firewall/iptables-1.4.21": {"EAPI": "5", "SLOT": "0"},
            "sys-apps/iproute2-3.12.0": {
                "EAPI": "5",
                "RDEPEND": "net-firewall/iptables:=",
            },
        }

        installed = {
            "net-firewall/iptables-1.4.21::overlay": {"EAPI": "5", "SLOT": "0/10"},
            "sys-apps/iproute2-3.12.0": {
                "EAPI": "5",
                "RDEPEND": "net-firewall/iptables:0/10=",
            },
        }

        world = ["sys-apps/iproute2"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--deep": True, "--update": True, "--verbose": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictMultiRepoUpdates(self):
        """
        Bug 508236 (similar to testSlotConflictMultiRepo)
        Different repositories contain the same cpv with different sub-slots for
        a slot operator child. For both the installed version and an updated version.

        """
        ebuilds = {
            "net-firewall/iptables-1.4.21::overlay": {"EAPI": "5", "SLOT": "0/10"},
            "net-firewall/iptables-1.4.21-r1::overlay": {"EAPI": "5", "SLOT": "0/10"},
            "sys-apps/iproute2-3.11.0::overlay": {
                "EAPI": "5",
                "RDEPEND": "net-firewall/iptables:=",
            },
            "net-firewall/iptables-1.4.21": {"EAPI": "5", "SLOT": "0"},
            "net-firewall/iptables-1.4.21-r1": {"EAPI": "5", "SLOT": "0"},
            "sys-apps/iproute2-3.12.0": {
                "EAPI": "5",
                "RDEPEND": "net-firewall/iptables:=",
            },
        }

        installed = {
            "net-firewall/iptables-1.4.21::overlay": {"EAPI": "5", "SLOT": "0/10"},
            "sys-apps/iproute2-3.12.0": {
                "EAPI": "5",
                "RDEPEND": "net-firewall/iptables:0/10=",
            },
        }

        world = ["sys-apps/iproute2"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--deep": True, "--update": True, "--verbose": True},
                success=True,
                mergelist=["net-firewall/iptables-1.4.21-r1::overlay"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSlotConflictRebuildGolang(self):

        ebuilds = {
            "dev-lang/go-1.14.7": {"EAPI": "7", "SLOT": "0/1.14.7"},
            "dev-lang/go-1.15": {"EAPI": "7", "SLOT": "0/1.15"},
            "net-p2p/syncthing-1.3.4-r1": {
                "EAPI": "7",
                "BDEPEND": "=dev-lang/go-1.14* >=dev-lang/go-1.12",
            },
        }

        installed = {
            "dev-lang/go-1.14.7": {"EAPI": "7", "SLOT": "0/1.14.7"},
            "net-p2p/syncthing-1.3.4-r1": {
                "EAPI": "7",
                "BDEPEND": "=dev-lang/go-1.14* >=dev-lang/go-1.12",
            },
        }

        world = ["dev-lang/go", "net-p2p/syncthing"]

        test_cases = (
            # Demonstrate an unwanted dev-lang/go rebuild triggered by a missed
            # update due to a slot conflict (bug #439688).
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()
